from typing import Optional

import torch

from tensordict.tensordict import TensorDict
from torchrl.data import (
    BoundedTensorSpec,
    CompositeSpec,
    UnboundedContinuousTensorSpec,
    UnboundedDiscreteTensorSpec,
)

from rl4co.data.utils import load_npz_to_tensordict
from rl4co.envs.common.base import RL4COEnvBase
from rl4co.utils.ops import gather_by_index, get_tour_length
from rl4co.utils.pylogger import get_pylogger

log = get_pylogger(__name__)


# From Kool et al. 2019, Hottung et al. 2022, Kim et al. 2023
CAPACITIES = {
    10: 20.0,
    15: 25.0,
    20: 30.0,
    30: 33.0,
    40: 37.0,
    50: 40.0,
    60: 43.0,
    75: 45.0,
    100: 50.0,
    125: 55.0,
    150: 60.0,
    200: 70.0,
    500: 100.0,
    1000: 150.0,
}


class CVRPEnv(RL4COEnvBase):
    """Capacitated Vehicle Routing Problem (CVRP) environment.
    At each step, the agent chooses a customer to visit depending on the current location and the remaining capacity.
    When the agent visits a customer, the remaining capacity is updated. If the remaining capacity is not enough to
    visit any customer, the agent must go back to the depot. The reward is the -infinite unless the agent visits all the cities.
    In that case, the reward is (-)length of the path: maximizing the reward is equivalent to minimizing the path length.

    Args:
        num_loc: number of locations (cities) in the VRP, without the depot. (e.g. 10 means 10 locs + 1 depot)
        min_loc: minimum value for the location coordinates
        max_loc: maximum value for the location coordinates
        min_demand: minimum value for the demand of each customer
        max_demand: maximum value for the demand of each customer
        vehicle_capacity: capacity of the vehicle
        capacity: capacity of the vehicle
        td_params: parameters of the environment
    """

    name = "cvrp"

    def __init__(
        self,
        num_loc: int = 20,
        min_loc: float = 0,
        max_loc: float = 1,
        min_demand: float = 1,
        max_demand: float = 10,
        vehicle_capacity: float = 1.0,
        capacity: float = None,
        td_params: TensorDict = None,
        **kwargs,
    ):
        super().__init__(**kwargs)
        self.num_loc = num_loc
        self.min_loc = min_loc
        self.max_loc = max_loc
        self.min_demand = min_demand
        self.max_demand = max_demand
        self.capacity = CAPACITIES.get(num_loc, None) if capacity is None else capacity
        if self.capacity is None:
            raise ValueError(
                f"Capacity for {num_loc} locations is not defined. Please provide a capacity manually."
            )
        self.vehicle_capacity = vehicle_capacity
        self._make_spec(td_params)

    def _step(self, td: TensorDict) -> TensorDict:
        current_node = td["action"][:, None]  # Add dimension for step
        n_loc = td["demand"].size(-1)  # Excludes depot

        # Not selected_demand is demand of first node (by clamp) so incorrect for nodes that visit depot!
        selected_demand = gather_by_index(
            td["demand"], torch.clamp(current_node - 1, 0, n_loc - 1), squeeze=False
        )

        # Increase capacity if depot is not visited, otherwise set to 0
        used_capacity = (td["used_capacity"] + selected_demand) * (
            current_node != 0
        ).float()

        # Note: here we do not subtract one as we have to scatter so the first column allows scattering depot
        # Add one dimension since we write a single value
        visited = td["visited"].scatter(-1, current_node[..., None], 1)

        # SECTION: get done
        done = visited.sum(-1) == visited.size(-1)
        reward = torch.ones_like(done) * float("-inf")

        td_step = TensorDict(
            {
                "next": {
                    "locs": td["locs"],
                    "demand": td["demand"],
                    "current_node": current_node,
                    "used_capacity": used_capacity,
                    "vehicle_capacity": td["vehicle_capacity"],
                    "visited": visited,
                    "reward": reward,
                    "done": done,
                }
            },
            td.shape,
        )
        td_step["next"].set("action_mask", self.get_action_mask(td_step["next"]))
        return td_step

    def _reset(
        self,
        td: Optional[TensorDict] = None,
        batch_size: Optional[list] = None,
    ) -> TensorDict:
        if batch_size is None:
            batch_size = self.batch_size if td is None else td["locs"].shape[:-2]

        if td is None or td.is_empty():
            td = self.generate_data(batch_size=batch_size)

        self.device = td.device

        # Create reset TensorDict
        td_reset = TensorDict(
            {
                "locs": torch.cat((td["depot"][:, None, :], td["locs"]), -2),
                "demand": td["demand"],
                "current_node": torch.zeros(
                    *batch_size, 1, dtype=torch.long, device=self.device
                ),
                "used_capacity": torch.zeros((*batch_size, 1), device=self.device),
                "vehicle_capacity": torch.full(
                    (*batch_size, 1), self.vehicle_capacity, device=self.device
                ),
                "visited": torch.zeros(
                    (*batch_size, 1, td["locs"].shape[-2] + 1),
                    dtype=torch.uint8,
                    device=self.device,
                ),
            },
            batch_size=batch_size,
        )
        td_reset.set("action_mask", self.get_action_mask(td_reset))
        return td_reset

    @staticmethod
    def get_action_mask(td: TensorDict) -> torch.Tensor:
        # For demand steps_dim is inserted by indexing with id, for used_capacity insert node dim for broadcasting
        exceeds_cap = td["demand"][:, None, :] + td["used_capacity"][..., None] > 1.0

        # Nodes that cannot be visited are already visited or too much demand to be served now
        mask_loc = td["visited"][..., 1:].to(exceeds_cap.dtype) | exceeds_cap

        # Cannot visit the depot if just visited and still unserved nodes
        mask_depot = (td["current_node"] == 0) & ((mask_loc == 0).int().sum(-1) > 0)
        return ~torch.cat((mask_depot[..., None], mask_loc), -1).squeeze(-2)

    def get_reward(self, td: TensorDict, actions: TensorDict) -> TensorDict:
        # Check that the solution is valid
        if self.check_solution:
            self.check_solution_validity(td, actions)

        # Gather dataset in order of tour
        depot = td["locs"][..., 0:1, :]
        locs_ordered = torch.cat([depot, gather_by_index(td["locs"], actions)], dim=1)
        return -get_tour_length(locs_ordered)

    @staticmethod
    def check_solution_validity(td: TensorDict, actions: torch.Tensor):
        """Check that solution is valid: nodes are not visited twice except depot and capacity is not exceeded"""
        # Check if tour is valid, i.e. contain 0 to n-1
        batch_size, graph_size = td["demand"].size()
        sorted_pi = actions.data.sort(1)[0]

        # Sorting it should give all zeros at front and then 1...n
        assert (
            torch.arange(1, graph_size + 1, out=sorted_pi.data.new())
            .view(1, -1)
            .expand(batch_size, graph_size)
            == sorted_pi[:, -graph_size:]
        ).all() and (sorted_pi[:, :-graph_size] == 0).all(), "Invalid tour"

        # Visiting depot resets capacity so we add demand = -capacity (we make sure it does not become negative)
        demand_with_depot = torch.cat((-td["vehicle_capacity"], td["demand"]), 1)
        d = demand_with_depot.gather(1, actions)

        used_cap = torch.zeros_like(td["demand"][:, 0])
        for i in range(actions.size(1)):
            used_cap += d[
                :, i
            ]  # This will reset/make capacity negative if i == 0, e.g. depot visited
            # Cannot use less than 0
            used_cap[used_cap < 0] = 0
            assert (
                used_cap <= td["vehicle_capacity"] + 1e-5
            ).all(), "Used more than capacity"

    def generate_data(self, batch_size) -> TensorDict:
        # Batch size input check
        batch_size = [batch_size] if isinstance(batch_size, int) else batch_size

        # Initialize the locations (including the depot which is always the first node)
        locs_with_depot = (
            torch.FloatTensor(*batch_size, self.num_loc + 1, 2)
            .uniform_(self.min_loc, self.max_loc)
            .to(self.device)
        )

        # Initialize the demand for nodes except the depot
        # Demand sampling Following Kool et al. (2019)
        # Generates a slightly different distribution than using torch.randint
        demand = (
            (
                torch.FloatTensor(*batch_size, self.num_loc)
                .uniform_(self.min_demand - 1, self.max_demand - 1)
                .int()
                + 1
            )
            .float()
            .to(self.device)
        )

        # Support for heterogeneous capacity if provided
        if not isinstance(self.capacity, torch.Tensor):
            capacity = torch.full((*batch_size,), self.capacity, device=self.device)
        else:
            capacity = self.capacity

        return TensorDict(
            {
                "locs": locs_with_depot[..., 1:, :],
                "depot": locs_with_depot[..., 0, :],
                "demand": demand / CAPACITIES[self.num_loc],
                "capacity": capacity,
            },
            batch_size=batch_size,
        )

    @staticmethod
    def load_data(fpath, batch_size=[]):
        """Dataset loading from file
        Normalize demand by capacity to be in [0, 1]
        """
        td_load = load_npz_to_tensordict(fpath)
        td_load.set_("demand", td_load["demand"] / td_load["capacity"][:, None])
        return td_load

    def _make_spec(self, td_params: TensorDict):
        """Make the observation and action specs from the parameters."""
        self.observation_spec = CompositeSpec(
            locs=BoundedTensorSpec(
                minimum=self.min_loc,
                maximum=self.max_loc,
                shape=(self.num_loc + 1, 2),
                dtype=torch.float32,
            ),
            current_node=UnboundedDiscreteTensorSpec(
                shape=(1),
                dtype=torch.int64,
            ),
            demand=BoundedTensorSpec(
                minimum=-self.capacity,
                maximum=self.max_demand,
                shape=(self.num_loc, 1),  # demand is only for customers
                dtype=torch.float32,
            ),
            action_mask=UnboundedDiscreteTensorSpec(
                shape=(self.num_loc + 1, 1),
                dtype=torch.bool,
            ),
            shape=(),
        )
        self.input_spec = self.observation_spec.clone()
        self.action_spec = BoundedTensorSpec(
            shape=(1,),
            dtype=torch.int64,
            minimum=0,
            maximum=self.num_loc + 1,
        )
        self.reward_spec = UnboundedContinuousTensorSpec(shape=(1,))
        self.done_spec = UnboundedDiscreteTensorSpec(shape=(1,), dtype=torch.bool)

    @staticmethod
    def render(td: TensorDict, actions=None, ax=None):
        import matplotlib.pyplot as plt
        import numpy as np

        from matplotlib import cm, colormaps

        num_routine = (actions == 0).sum().item() + 2
        base = colormaps["nipy_spectral"]
        color_list = base(np.linspace(0, 1, num_routine))
        cmap_name = base.name + str(num_routine)
        out = base.from_list(cmap_name, color_list, num_routine)

        if ax is None:
            # Create a plot of the nodes
            _, ax = plt.subplots()

        td = td.detach().cpu()

        if actions is None:
            actions = td.get("action", None)

        # if batch_size greater than 0 , we need to select the first batch element
        if td.batch_size != torch.Size([]):
            td = td[0]
            actions = actions[0]

        locs = td["locs"]
        scale = CAPACITIES.get(td["locs"].size(-2) - 1, 1)
        demands = td["demand"] * scale

        # add the depot at the first action and the end action
        actions = torch.cat([torch.tensor([0]), actions, torch.tensor([0])])

        # gather locs in order of action if available
        if actions is None:
            log.warning("No action in TensorDict, rendering unsorted locs")
        else:
            locs = locs

        # Cat the first node to the end to complete the tour
        x, y = locs[:, 0], locs[:, 1]

        # plot depot
        ax.scatter(
            locs[0, 0],
            locs[0, 1],
            edgecolors=cm.Set2(2),
            facecolors="none",
            s=100,
            linewidths=2,
            marker="s",
            alpha=1,
        )

        # plot visited nodes
        ax.scatter(
            x[1:],
            y[1:],
            edgecolors=cm.Set2(0),
            facecolors="none",
            s=50,
            linewidths=2,
            marker="o",
            alpha=1,
        )

        # plot demand bars
        for node_idx in range(1, len(locs)):
            ax.add_patch(
                plt.Rectangle(
                    (locs[node_idx, 0] - 0.005, locs[node_idx, 1] + 0.015),
                    0.01,
                    demands[node_idx - 1] / (scale * 10),
                    edgecolor=cm.Set2(0),
                    facecolor=cm.Set2(0),
                    fill=True,
                )
            )

        # text demand
        for node_idx in range(1, len(locs)):
            ax.text(
                locs[node_idx, 0],
                locs[node_idx, 1] - 0.025,
                f"{demands[node_idx-1].item():.2f}",
                horizontalalignment="center",
                verticalalignment="top",
                fontsize=10,
                color=cm.Set2(0),
            )

        # text depot
        ax.text(
            locs[0, 0],
            locs[0, 1] - 0.025,
            "Depot",
            horizontalalignment="center",
            verticalalignment="top",
            fontsize=10,
            color=cm.Set2(2),
        )

        # plot actions
        color_idx = 0
        for action_idx in range(len(actions) - 1):
            if actions[action_idx] == 0:
                color_idx += 1
            from_loc = locs[actions[action_idx]]
            to_loc = locs[actions[action_idx + 1]]
            ax.plot(
                [from_loc[0], to_loc[0]],
                [from_loc[1], to_loc[1]],
                color=out(color_idx),
                lw=1,
            )
            ax.annotate(
                "",
                xy=(to_loc[0], to_loc[1]),
                xytext=(from_loc[0], from_loc[1]),
                arrowprops=dict(arrowstyle="-|>", color=out(color_idx)),
                size=15,
                annotation_clip=False,
            )

        # Setup limits and show
        ax.set_xlim(-0.05, 1.05)
        ax.set_ylim(-0.05, 1.05)
        plt.show()



import pytest

from rl4co.models import AutoregressivePolicy, PointerNetworkPolicy
from rl4co.utils.test_utils import generate_env_data


# Main autorergressive policy: rollout over multiple envs since it is the base
@pytest.mark.parametrize(
    "env_name", ["tsp", "cvrp", "sdvrp", "mtsp", "op", "pctsp", "spctsp", "dpp", "mdpp"]
)
def test_base_policy(env_name, size=20, batch_size=2):
    env, x = generate_env_data(env_name, size, batch_size)
    td = env.reset(x)
    policy = AutoregressivePolicy(env.name)
    out = policy(td, env, decode_type="greedy")
    assert out["reward"].shape == (batch_size,)


@pytest.mark.parametrize("env_name", ["tsp", "cvrp", "pctsp", "spctsp"])
def test_base_policy_multistart(env_name, size=20, batch_size=2):
    env, x = generate_env_data(env_name, size, batch_size)
    td = env.reset(x)
    policy = AutoregressivePolicy(env.name)
    out = policy(td, env, decode_type="greedy_multistart", num_starts=size)
    assert out["reward"].shape == (
        batch_size * size,
    )  # to evaluate, we could just unbatchify


def test_pointer_network(size=20, batch_size=2):
    env, x = generate_env_data("tsp", size, batch_size)
    td = env.reset(x)
    policy = PointerNetworkPolicy(env.name)
    out = policy(td, env, decode_type="greedy")
    assert out["reward"].shape == (batch_size,)



import torch

from rl4co.utils import get_pylogger

log = get_pylogger(__name__)


def get_log_likelihood(log_p, actions, mask, return_sum: bool = True):
    """Get log likelihood of selected actions"""

    log_p = log_p.gather(2, actions.unsqueeze(-1)).squeeze(-1)

    # Optional: mask out actions irrelevant to objective so they do not get reinforced
    if mask is not None:
        log_p[~mask] = 0

    assert (
        log_p > -1000
    ).data.all(), "Logprobs should not be -inf, check sampling procedure!"

    # Calculate log_likelihood
    if return_sum:
        return log_p.sum(1)  # [batch]
    else:
        return log_p  # [batch, decode_len]


def decode_probs(probs, mask, decode_type="sampling"):
    """Decode probabilities to select actions with mask"""

    assert (probs == probs).all(), "Probs should not contain any nans"

    if "greedy" in decode_type:
        _, selected = probs.max(1)
        assert not mask.gather(
            1, selected.unsqueeze(-1)
        ).data.any(), "Decode greedy: infeasible action has maximum probability"

    elif "sampling" in decode_type:
        selected = torch.multinomial(probs, 1).squeeze(1)

        while mask.gather(1, selected.unsqueeze(-1)).data.any():
            log.info("Sampled bad values, resampling!")
            selected = probs.multinomial(1).squeeze(1)

    else:
        assert False, "Unknown decode type: {}".format(decode_type)
    return selected


def random_policy(td):
    """Helper function to select a random action from available actions"""
    action = torch.multinomial(td["action_mask"].float(), 1).squeeze(-1)
    td.set("action", action)
    return td


def rollout(env, td, policy):
    """Helper function to rollout a policy. Currently, TorchRL does not allow to step
    over envs when done with `env.rollout()`. We need this because for environements that complete at different steps.
    """
    actions = []
    while not td["done"].all():
        td = policy(td)
        actions.append(td["action"])
        td = env.step(td)["next"]
    return (
        env.get_reward(td, torch.stack(actions, dim=1)),
        td,
        torch.stack(actions, dim=1),
    )



import torch
import torch.nn as nn


def env_init_embedding(env_name: str, config: dict) -> nn.Module:
    """Get environment initial embedding. The init embedding is used to initialize the
    general embedding of the problem nodes without any solution information.
    Consists of a linear layer that projects the node features to the embedding space.

    Args:
        env: Environment or its name.
        config: A dictionary of configuration options for the environment.
    """
    embedding_registry = {
        "tsp": TSPInitEmbedding,
        "atsp": TSPInitEmbedding,
        "cvrp": VRPInitEmbedding,
        "sdvrp": VRPInitEmbedding,
        "pctsp": PCTSPInitEmbedding,
        "spctsp": PCTSPInitEmbedding,
        "op": OPInitEmbedding,
        "dpp": DPPInitEmbedding,
        "mdpp": MDPPInitEmbedding,
        "pdp": PDPInitEmbedding,
        "mtsp": MTSPInitEmbedding,
    }

    if env_name not in embedding_registry:
        raise ValueError(
            f"Unknown environment name '{env_name}'. Available init embeddings: {embedding_registry.keys()}"
        )

    return embedding_registry[env_name](**config)


class TSPInitEmbedding(nn.Module):
    """Initial embedding for the Traveling Salesman Problems (TSP).
    Embed the following node features to the embedding space:
        - locs: x, y coordinates of the cities
    """

    def __init__(self, embedding_dim):
        super(TSPInitEmbedding, self).__init__()
        node_dim = 2  # x, y
        self.init_embed = nn.Linear(node_dim, embedding_dim)

    def forward(self, td):
        out = self.init_embed(td["locs"])
        return out


class VRPInitEmbedding(nn.Module):
    """Initial embedding for the Vehicle Routing Problems (VRP).
    Embed the following node features to the embedding space:
        - locs: x, y coordinates of the nodes (depot and customers separately)
        - demand: demand of the customers
    """

    def __init__(self, embedding_dim):
        super(VRPInitEmbedding, self).__init__()
        node_dim = 3  # x, y, demand
        self.init_embed = nn.Linear(node_dim, embedding_dim)
        self.init_embed_depot = nn.Linear(2, embedding_dim)  # depot embedding

    def forward(self, td):
        # [batch, 1, 2]-> [batch, 1, embedding_dim]
        depot, cities = td["locs"][:, :1, :], td["locs"][:, 1:, :]
        depot_embedding = self.init_embed_depot(depot)
        # [batch, n_city, 2, batch, n_city, 1]  -> [batch, n_city, embedding_dim]
        node_embeddings = self.init_embed(
            torch.cat((cities, td["demand"][..., None]), -1)
        )
        # [batch, n_city+1, embedding_dim]
        out = torch.cat((depot_embedding, node_embeddings), -2)
        return out


class PCTSPInitEmbedding(nn.Module):
    """Initial embedding for the Prize Collecting Traveling Salesman Problems (PCTSP).
    Embed the following node features to the embedding space:
        - locs: x, y coordinates of the nodes (depot and customers separately)
        - expected_prize: expected prize for visiting the customers.
            In PCTSP, this is the actual prize. In SPCTSP, this is the expected prize.
        - penalty: penalty for not visiting the customers
    """

    def __init__(self, embedding_dim):
        super(PCTSPInitEmbedding, self).__init__()
        node_dim = 4  # x, y, prize, penalty
        self.init_embed = nn.Linear(node_dim, embedding_dim)
        self.init_embed_depot = nn.Linear(2, embedding_dim)

    def forward(self, td):
        depot, cities = td["locs"][:, :1, :], td["locs"][:, 1:, :]
        depot_embedding = self.init_embed_depot(depot)
        node_embeddings = self.init_embed(
            torch.cat(
                (
                    cities,
                    td["expected_prize"][..., None],
                    td["penalty"][..., 1:, None],
                ),
                -1,
            )
        )
        # batch, n_city+1, embedding_dim
        out = torch.cat((depot_embedding, node_embeddings), -2)
        return out


class OPInitEmbedding(nn.Module):
    """Initial embedding for the Orienteering Problems (OP).
    Embed the following node features to the embedding space:
        - locs: x, y coordinates of the nodes (depot and customers separately)
        - prize: prize for visiting the customers
    """

    def __init__(self, embedding_dim):
        super(OPInitEmbedding, self).__init__()
        node_dim = 3  # x, y, prize
        self.init_embed = nn.Linear(node_dim, embedding_dim)
        self.init_embed_depot = nn.Linear(2, embedding_dim)  # depot embedding

    def forward(self, td):
        depot, cities = td["locs"][:, :1, :], td["locs"][:, 1:, :]
        depot_embedding = self.init_embed_depot(depot)
        node_embeddings = self.init_embed(
            torch.cat(
                (
                    cities,
                    td["prize"][..., 1:, None],  # exclude depot
                ),
                -1,
            )
        )
        out = torch.cat((depot_embedding, node_embeddings), -2)
        return out


class DPPInitEmbedding(nn.Module):
    """Initial embedding for the Decap Placement Problem (DPP), EDA (electronic design automation).
    Embed the following node features to the embedding space:
        - locs: x, y coordinates of the nodes (cells)
        - probe: index of the (single) probe cell. We embed the euclidean distance from the probe to all cells.
    """

    def __init__(self, embedding_dim):
        super(DPPInitEmbedding, self).__init__()
        node_dim = 2  # x, y
        self.init_embed = nn.Linear(node_dim, embedding_dim // 2)  # locs
        self.init_embed_probe = nn.Linear(1, embedding_dim // 2)  # probe

    def forward(self, td):
        node_embeddings = self.init_embed(td["locs"])
        probe_embedding = self.init_embed_probe(
            self._distance_probe(td["locs"], td["probe"])
        )
        return torch.cat([node_embeddings, probe_embedding], -1)

    def _distance_probe(self, locs, probe):
        # Euclidean distance from probe to all locations
        probe_loc = torch.gather(locs, 1, probe.unsqueeze(-1).expand(-1, -1, 2))
        return torch.norm(locs - probe_loc, dim=-1).unsqueeze(-1)


class MDPPInitEmbedding(nn.Module):
    """Initial embedding for the Multi-port Placement Problem (MDPP), EDA (electronic design automation).
    Embed the following node features to the embedding space:
        - locs: x, y coordinates of the nodes (cells)
        - probe: indexes of the probe cells (multiple). We embed the euclidean distance of each cell to the closest probe.
    """

    def __init__(self, embedding_dim):
        super(MDPPInitEmbedding, self).__init__()
        node_dim = 2  # x, y
        self.init_embed = nn.Linear(node_dim, embedding_dim)  # locs
        self.init_embed_probe_distance = nn.Linear(1, embedding_dim)  # probe_distance
        self.project_out = nn.Linear(embedding_dim * 2, embedding_dim)

    def forward(self, td):
        probes = td["probe"]
        locs = td["locs"]
        node_embeddings = self.init_embed(locs)

        # Get the shortest distance from any probe
        dist = torch.cdist(locs, locs, p=2)
        dist[~probes] = float("inf")
        min_dist, _ = torch.min(dist, dim=1)
        min_probe_dist_embedding = self.init_embed_probe_distance(min_dist[..., None])

        return self.project_out(
            torch.cat([node_embeddings, min_probe_dist_embedding], -1)
        )


class PDPInitEmbedding(nn.Module):
    """Initial embedding for the Pickup and Delivery Problem (PDP).
    Embed the following node features to the embedding space:
        - locs: x, y coordinates of the nodes (depot, pickups and deliveries separately)
           Note that pickups and deliveries are interleaved in the input.
    """

    def __init__(self, embedding_dim):
        super(PDPInitEmbedding, self).__init__()
        node_dim = 2  # x, y
        self.init_embed_depot = nn.Linear(2, embedding_dim)
        self.init_embed_pick = nn.Linear(node_dim * 2, embedding_dim)
        self.init_embed_delivery = nn.Linear(node_dim, embedding_dim)

    def forward(self, td):
        depot, locs = td["locs"][..., 0:1, :], td["locs"][..., 1:, :]
        num_locs = locs.size(-2)
        pick_feats = torch.cat(
            [locs[:, : num_locs // 2, :], locs[:, num_locs // 2 :, :]], -1
        )  # [batch_size, graph_size//2, 4]
        delivery_feats = locs[:, num_locs // 2 :, :]  # [batch_size, graph_size//2, 2]
        depot_embeddings = self.init_embed_depot(depot)
        pick_embeddings = self.init_embed_pick(pick_feats)
        delivery_embeddings = self.init_embed_delivery(delivery_feats)
        # concatenate on graph size dimension
        return torch.cat([depot_embeddings, pick_embeddings, delivery_embeddings], -2)


class MTSPInitEmbedding(nn.Module):
    """Initial embedding for the Multiple Traveling Salesman Problem (mTSP).
    Embed the following node features to the embedding space:
        - locs: x, y coordinates of the nodes (depot, cities)
    """

    def __init__(self, embedding_dim):
        """NOTE: new made by Fede. May need to be checked"""
        super(MTSPInitEmbedding, self).__init__()
        node_dim = 2  # x, y
        self.init_embed = nn.Linear(node_dim, embedding_dim)
        self.init_embed_depot = nn.Linear(2, embedding_dim)  # depot embedding

    def forward(self, td):
        depot_embedding = self.init_embed_depot(td["locs"][..., 0:1, :])
        node_embedding = self.init_embed(td["locs"][..., 1:, :])
        return torch.cat([depot_embedding, node_embedding], -2)



from dataclasses import dataclass
from typing import Tuple, Union

import torch
import torch.nn as nn

from einops import rearrange
from tensordict import TensorDict
from torch import Tensor

from rl4co.envs import RL4COEnvBase, get_env
from rl4co.models.nn.attention import LogitAttention
from rl4co.models.nn.env_embeddings import env_context_embedding, env_dynamic_embedding
from rl4co.models.nn.utils import decode_probs
from rl4co.utils.ops import batchify, select_start_nodes, unbatchify


@dataclass
class PrecomputedCache:
    node_embeddings: Tensor
    graph_context: Union[Tensor, float]
    glimpse_key: Tensor
    glimpse_val: Tensor
    logit_key: Tensor


class AutoregressiveDecoder(nn.Module):
    """Auto-regressive decoder for constructing solutions for combinatorial optimization problems.
    Given the environment state and the embeddings, compute the logits and sample actions autoregressively until
    all the environments in the batch have reached a terminal state.
    We additionally include support for multi-starts as it is more efficient to do so in the decoder as we can
    natively perform the attention computation.

    Note:
        There are major differences between this decoding and most RL problems. The most important one is
        that reward is not defined for partial solutions, hence we have to wait for the environment to reach a terminal
        state before we can compute the reward with `env.get_reward()`.

    Warning:
        We suppose environments in the `done` state are still available for sampling. This is because in NCO we need to
        wait for all the environments to reach a terminal state before we can stop the decoding process. This is in
        contrast with the TorchRL framework (at the moment) where the `env.rollout` function automatically resets.
        You may follow tighter integration with TorchRL here: https://github.com/kaist-silab/rl4co/issues/72.

    Args:
        env_name: environment name to solve
        embedding_dim: Dimension of the embeddings
        num_heads: Number of heads for the attention
        use_graph_context: Whether to use the initial graph context to modify the query
    """

    def __init__(
        self,
        env_name: str,
        embedding_dim: int,
        num_heads: int,
        use_graph_context: bool = True,
        **logit_attn_kwargs,
    ):
        super().__init__()

        self.env_name = env_name
        self.embedding_dim = embedding_dim
        self.num_heads = num_heads

        assert embedding_dim % num_heads == 0

        self.context_embedding = env_context_embedding(
            self.env_name, {"embedding_dim": embedding_dim}
        )
        self.dynamic_embedding = env_dynamic_embedding(
            self.env_name, {"embedding_dim": embedding_dim}
        )
        self.use_graph_context = use_graph_context

        # For each node we compute (glimpse key, glimpse value, logit key) so 3 * embedding_dim
        self.project_node_embeddings = nn.Linear(
            embedding_dim, 3 * embedding_dim, bias=False
        )
        self.project_fixed_context = nn.Linear(embedding_dim, embedding_dim, bias=False)

        # MHA
        self.logit_attention = LogitAttention(
            embedding_dim, num_heads, **logit_attn_kwargs
        )

    def forward(
        self,
        td: TensorDict,
        embeddings: Tensor,
        env: Union[str, RL4COEnvBase] = None,
        decode_type: str = "sampling",
        num_starts: int = None,
        softmax_temp: float = None,
        calc_reward: bool = True,
    ) -> Tuple[Tensor, Tensor, TensorDict]:
        """Forward pass of the decoder
        Given the environment state and the pre-computed embeddings, compute the logits and sample actions

        Args:
            td: Input TensorDict containing the environment state
            embeddings: Precomputed embeddings for the nodes
            env: Environment to use for decoding. If None, the environment is instantiated from `env_name`. Note that
                it is more efficient to pass an already instantiated environment each time for fine-grained control
            decode_type: Type of decoding to use. Can be one of:
                - "sampling": sample from the logits
                - "greedy": take the argmax of the logits
                - "multistart_sampling": sample as sampling, but with multi-start decoding
                - "multistart_greedy": sample as greedy, but with multi-start decoding
            num_starts: Number of multi-starts to use. If None, no multi-start decoding is used
            softmax_temp: Temperature for the softmax. If None, default softmax is used from the `LogitAttention` module
            calc_reward: Whether to calculate the reward for the decoded sequence

        Returns:
            outputs: Tensor of shape (batch_size, seq_len, num_nodes) containing the logits
            actions: Tensor of shape (batch_size, seq_len) containing the sampled actions
            td: TensorDict containing the environment state after decoding
        """

        # Greedy multi-start decoding if num_starts > 1
        num_starts = 0 if num_starts is None else num_starts
        assert not (
            "multistart" in decode_type and num_starts <= 1
        ), "Multi-start decoding requires `num_starts` > 1"

        # Compute keys, values for the glimpse and keys for the logits once as they can be reused in every step
        cached_embeds = self._precompute_cache(embeddings, num_starts=num_starts)

        # Collect outputs
        outputs = []
        actions = []

        # Instantiate environment if needed
        if isinstance(env, str):
            env_name = self.env_name if env is None else env
            env = get_env(env_name)

        # Multi-start decoding: first action is chosen by ad-hoc node selection
        if num_starts > 1 or "multistart" in decode_type:
            action = select_start_nodes(td, num_starts, env)

            # Expand td to batch_size * num_starts
            td = batchify(td, num_starts)

            td.set("action", action)
            td = env.step(td)["next"]
            log_p = torch.zeros_like(
                td["action_mask"], device=td.device
            )  # first log_p is 0, so p = log_p.exp() = 1

            outputs.append(log_p)
            actions.append(action)

        # Main decoding
        while not td["done"].all():
            log_p, mask = self._get_log_p(cached_embeds, td, softmax_temp, num_starts)

            # Select the indices of the next nodes in the sequences, result (batch_size) long
            action = decode_probs(log_p.exp(), mask, decode_type=decode_type)

            td.set("action", action)
            td = env.step(td)["next"]

            # Collect output of step
            outputs.append(log_p)
            actions.append(action)

        outputs, actions = torch.stack(outputs, 1), torch.stack(actions, 1)
        if calc_reward:
            td.set("reward", env.get_reward(td, actions))

        return outputs, actions, td

    def _precompute_cache(self, embeddings: Tensor, num_starts: int = 0):
        """Compute the cached embeddings for the attention

        Args:
            embeddings: Precomputed embeddings for the nodes
            num_starts: Number of multi-starts to use. If 0, no multi-start decoding is used
        """

        # The projection of the node embeddings for the attention is calculated once up front
        (
            glimpse_key_fixed,
            glimpse_val_fixed,
            logit_key_fixed,
        ) = self.project_node_embeddings(embeddings).chunk(3, dim=-1)

        # Optionally disable the graph context from the initial embedding as done in POMO
        if self.use_graph_context:
            graph_context = unbatchify(
                batchify(self.project_fixed_context(embeddings.mean(1)), num_starts),
                num_starts,
            )
        else:
            graph_context = 0

        # Organize in a dataclass for easy access
        cached_embeds = PrecomputedCache(
            node_embeddings=embeddings,
            graph_context=graph_context,
            glimpse_key=glimpse_key_fixed,
            glimpse_val=glimpse_val_fixed,
            logit_key=logit_key_fixed,
        )

        return cached_embeds

    def _get_log_p(
        self,
        cached: PrecomputedCache,
        td: TensorDict,
        softmax_temp: float = None,
        num_starts: int = 0,
    ):
        """Compute the log probabilities of the next actions given the current state

        Args:
            cache: Precomputed embeddings
            td: TensorDict with the current environment state
            softmax_temp: Temperature for the softmax
            num_starts: Number of starts for the multi-start decoding
        """

        # Unbatchify to [batch_size, num_starts, ...]. Has no effect if num_starts = 0
        td_unbatch = unbatchify(td, num_starts)
        step_context = self.context_embedding(cached.node_embeddings, td_unbatch)
        glimpse_q = step_context + cached.graph_context
        glimpse_q = glimpse_q.unsqueeze(1) if glimpse_q.ndim == 2 else glimpse_q

        # Compute keys and values for the nodes
        (
            glimpse_key_dynamic,
            glimpse_val_dynamic,
            logit_key_dynamic,
        ) = self.dynamic_embedding(td_unbatch)
        glimpse_k = cached.glimpse_key + glimpse_key_dynamic
        glimpse_v = cached.glimpse_val + glimpse_val_dynamic
        logit_k = cached.logit_key + logit_key_dynamic

        # Get the mask
        mask = ~td_unbatch["action_mask"]

        # Compute logits
        log_p = self.logit_attention(
            glimpse_q, glimpse_k, glimpse_v, logit_k, mask, softmax_temp
        )

        # Now we need to reshape the logits and log_p to [batch_size*num_starts, num_nodes]
        # Note that rearranging order is important here
        log_p = rearrange(log_p, "b s l -> (s b) l") if num_starts > 1 else log_p
        mask = rearrange(mask, "b s l -> (s b) l") if num_starts > 1 else mask
        return log_p, mask



from typing import Union

from rl4co.envs.common.base import RL4COEnvBase
from rl4co.models.rl import REINFORCE
from rl4co.models.rl.reinforce.baselines import REINFORCEBaseline
from rl4co.models.zoo.ptrnet.policy import PointerNetworkPolicy


class PointerNetwork(REINFORCE):
    """
    Pointer Network for neural combinatorial optimization based on REINFORCE
    Based on Vinyals et al. (2015) https://arxiv.org/abs/1506.03134
    Refactored from reference implementation: https://github.com/wouterkool/attention-learn-to-route
    Args:
        env: Environment to use for the algorithm
        policy: Policy to use for the algorithm
        baseline: REINFORCE baseline. Defaults to rollout (1 epoch of exponential, then greedy rollout baseline)
        policy_kwargs: Keyword arguments for policy
        baseline_kwargs: Keyword arguments for baseline
        **kwargs: Keyword arguments passed to the superclass
    """

    def __init__(
        self,
        env: RL4COEnvBase,
        policy: PointerNetworkPolicy = None,
        baseline: Union[REINFORCEBaseline, str] = "rollout",
        policy_kwargs={},
        baseline_kwargs={},
        **kwargs,
    ):
        self.policy = (
            PointerNetworkPolicy(self.env, **policy_kwargs) if policy is None else policy
        )
        super().__init__(env, policy, baseline, baseline_kwargs, **kwargs)



import inspect

import hydra
import torch.nn as nn


def group_parameters_for_optimizer(
    model, optimizer_cfg, bias_weight_decay=False, normalization_weight_decay=False
):
    """Set weight_decay=0.0 for parameters in model.no_weight_decay, for parameters with
    attribute _no_weight_decay==True, for bias parameters if bias_weight_decay==False, for
    normalization parameters if normalization_weight_decay==False
    """
    # Get the weight decay from the config, or from the default value of the optimizer constructor
    # if it's not specified in the config.
    if "weight_decay" in optimizer_cfg:
        weight_decay = optimizer_cfg.weight_decay
    else:
        # https://stackoverflow.com/questions/12627118/get-a-function-arguments-default-value
        signature = inspect.signature(hydra.utils.get_class(optimizer_cfg._target_))
        if "weight_decay" in signature.parameters:
            weight_decay = signature.parameters["weight_decay"].default
            if weight_decay is inspect.Parameter.empty:
                weight_decay = 0.0
        else:
            weight_decay = 0.0

    # If none of the parameters have weight decay anyway, and there are no parameters with special
    # optimization params
    if weight_decay == 0.0 and not any(hasattr(p, "_optim") for p in model.parameters()):
        return model.parameters()

    skip = model.no_weight_decay() if hasattr(model, "no_weight_decay") else set()
    skip_keywords = (
        model.no_weight_decay_keywords()
        if hasattr(model, "no_weight_decay_keywords")
        else set()
    )

    # Adapted from https://github.com/karpathy/minGPT/blob/master/mingpt/model.py#L134
    """
    This long function is unfortunately doing something very simple and is being very defensive:
    We are separating out all parameters of the model into two buckets: those that will experience
    weight decay for regularization and those that won't (biases, and layernorm/embedding weights).
    We are then returning the PyTorch optimizer object.
    """

    # separate out all parameters to those that will and won't experience regularizing weight decay
    decay = set()
    no_decay = set()
    special = set()
    whitelist_weight_modules = (nn.Linear,)
    blacklist_weight_modules = nn.Embedding
    if not normalization_weight_decay:
        blacklist_weight_modules += (
            nn.BatchNorm1d,
            nn.BatchNorm2d,
            nn.BatchNorm3d,
            nn.LazyBatchNorm1d,
            nn.LazyBatchNorm2d,
            nn.LazyBatchNorm3d,
            nn.GroupNorm,
            nn.SyncBatchNorm,
            nn.InstanceNorm1d,
            nn.InstanceNorm2d,
            nn.InstanceNorm3d,
            nn.LayerNorm,
            nn.LocalResponseNorm,
        )

    param_dict = {pn: p for pn, p in model.named_parameters() if p.requires_grad}
    for mn, m in model.named_modules():
        for pn, p in m.named_parameters():
            fpn = "%s.%s" % (mn, pn) if mn else pn  # full param name
            # In case of parameter sharing, some parameters show up here but are not in
            # param_dict.keys()
            if not p.requires_grad or fpn not in param_dict:
                continue  # frozen weights
            if hasattr(p, "_optim"):
                special.add(fpn)
            elif fpn in skip or any(
                skip_keyword in fpn for skip_keyword in skip_keywords
            ):
                no_decay.add(fpn)
            elif getattr(p, "_no_weight_decay", False):
                no_decay.add(fpn)
            elif not bias_weight_decay and pn.endswith("bias"):
                no_decay.add(fpn)
            elif pn.endswith("weight") and isinstance(m, whitelist_weight_modules):
                # weights of whitelist modules will be weight decayed
                decay.add(fpn)
            elif isinstance(m, blacklist_weight_modules):
                # weights of blacklist modules will NOT be weight decayed
                no_decay.add(fpn)

    # special case the position embedding parameter in the root GPT module as not decayed
    if "pos_emb" in param_dict:
        no_decay.add("pos_emb")

    decay |= param_dict.keys() - no_decay - special
    # validate that we considered every parameter
    inter_params = decay & no_decay
    union_params = decay | no_decay
    assert (
        len(inter_params) == 0
    ), f"Parameters {str(inter_params)} made it into both decay/no_decay sets!"
    assert (
        len(param_dict.keys() - special - union_params) == 0
    ), f"parameters {str(param_dict.keys() - union_params)}  were not separated into either decay/no_decay set!"

    if weight_decay == 0.0 or not no_decay:
        param_groups = [
            {
                "params": [param_dict[pn] for pn in sorted(list(no_decay | decay))],
                "weight_decay": weight_decay,
            }
        ]
    else:
        param_groups = [
            {
                "params": [param_dict[pn] for pn in sorted(list(decay))],
                "weight_decay": weight_decay,
            },
            {
                "params": [param_dict[pn] for pn in sorted(list(no_decay))],
                "weight_decay": 0.0,
            },
        ]
    # Add parameters with special hyperparameters
    # Unique dicts
    hps = [
        dict(s) for s in set(frozenset(param_dict[pn]._optim.items()) for pn in special)
    ]
    for hp in hps:
        params = [param_dict[pn] for pn in special if param_dict[pn]._optim == hp]
        param_groups.append({"params": params, **hp})

    return param_groups



import torch
import torch.nn as nn
import torch.nn.functional as F

from torch_geometric.data import Batch, Data
from torch_geometric.nn import GCNConv

from rl4co.utils.pylogger import get_pylogger

log = get_pylogger(__name__)


class GCNEncoder(nn.Module):
    """Graph Convolutional Network to encode embeddings with a series of GCN layers

    Args:
        embedding_dim: dimension of the embeddings
        num_nodes: number of nodes in the graph
        num_gcn_layer: number of GCN layers
        self_loop: whether to add self loop in the graph
        residual: whether to use residual connection
    """

    def __init__(
        self,
        embedding_dim: int,
        num_nodes: int,
        num_gcn_layer: int,
        self_loop: bool = False,
        residual: bool = True,
    ):
        super(GCNEncoder, self).__init__()

        # Generate edge index for a fully connected graph
        adj_matrix = torch.ones(num_nodes, num_nodes)
        if self_loop:
            adj_matrix.fill_diagonal_(0)  # No self-loops
        self.edge_index = torch.permute(torch.nonzero(adj_matrix), (1, 0))

        # Define the GCN layers
        self.gcn_layers = nn.ModuleList(
            [GCNConv(embedding_dim, embedding_dim) for _ in range(num_gcn_layer)]
        )

        # Record parameters
        self.residual = residual
        self.self_loop = self_loop

    def forward(self, x, node_feature, mask=None):
        """Forward pass of the GCN encoder

        Args:
            x: [batch_size, graph_size, embed_dim] initial embeddings to process
            node_feature: [batch_size, graph_size, embed_dim] node features, i.e. raw ones
            mask: [batch_size, graph_size] mask for valid nodes
        """

        assert mask is None, "Mask not yet supported!"
        # initial Embedding from features

        # Check to update the edge index with different number of node
        if node_feature.size(1) != self.edge_index.max().item() + 1:
            adj_matrix = torch.ones(x.size(1), x.size(1))
            if self.self_loop:
                adj_matrix.fill_diagonal_(0)
            edge_index = torch.permute(torch.nonzero(adj_matrix), (1, 0))
        else:
            edge_index = self.edge_index

        # Create the batched graph
        data_list = [Data(x=x, edge_index=edge_index) for x in node_feature]
        data_batch = Batch.from_data_list(data_list)

        # GCN process
        update_node_feature = data_batch.x
        edge_index = data_batch.edge_index
        for layer in self.gcn_layers[:-1]:
            update_node_feature = layer(update_node_feature, edge_index)
            update_node_feature = F.relu(update_node_feature)
            update_node_feature = F.dropout(update_node_feature, training=self.training)

        update_node_feature = self.gcn_layers[-1](update_node_feature, edge_index)

        # De-batch the graph
        input_size = node_feature.size()
        update_node_feature = update_node_feature.view(*input_size)

        # Residual
        update_node_feature = update_node_feature + node_feature

        return update_node_feature, node_feature


if __name__ == "__main__":
    from rl4co.envs import TSPEnv

    env = TSPEnv()
    model = GCNEncoder(
        env=env,
        embedding_dim=128,
        num_nodes=20,
        num_gcn_layer=3,
    )
    td = env.reset(batch_size=[32])
    update_node_feature, _ = model(td)
    print(update_node_feature.size())



from typing import Optional

import torch
import torch.nn.functional as F

from tensordict.tensordict import TensorDict
from torchrl.data import (
    BoundedTensorSpec,
    CompositeSpec,
    UnboundedContinuousTensorSpec,
    UnboundedDiscreteTensorSpec,
)

from rl4co.envs.common.base import RL4COEnvBase
from rl4co.utils.ops import gather_by_index, get_tour_length
from rl4co.utils.pylogger import get_pylogger

log = get_pylogger(__name__)

# For the penalty to make sense it should be not too large (in which case all nodes will be visited) nor too small
# so we want the objective term to be approximately equal to the length of the tour, which we estimate with half
# of the nodes by half of the tour length (which is very rough but similar to op)
# This means that the sum of penalties for all nodes will be approximately equal to the tour length (on average)
# The expected total (uniform) penalty of half of the nodes (since approx half will be visited by the constraint)
# is (n / 2) / 2 = n / 4 so divide by this means multiply by 4 / n,
# However instead of 4 we use penalty_factor (3 works well) so we can make them larger or smaller
MAX_LENGTHS = {20: 2.0, 50: 3.0, 100: 4.0}


class PCTSPEnv(RL4COEnvBase):
    """Prize-collecting TSP (PCTSP) environment.
    The goal is to collect as much prize as possible while minimizing the total travel cost.
    The environment is stochastic, the prize is only revealed when the node is visited.

    Args:
        num_loc: Number of locations
        min_loc: Minimum location value
        max_loc: Maximum location value
        penalty_factor: Penalty factor
        prize_required: Minimum prize required to visit a node
        check_solution: Set to False by default for small bug happening around 0.01% of the time (TODO: fix)
        td_params: Parameters of the environment
    """

    name = "pctsp"
    _stochastic = False

    def __init__(
        self,
        num_loc: int = 10,
        min_loc: float = 0,
        max_loc: float = 1,
        penalty_factor: float = 3,
        prize_required: float = 1,
        check_solution: bool = False,
        td_params: TensorDict = None,
        **kwargs,
    ):
        super().__init__(**kwargs)
        self.num_loc = num_loc
        self.min_loc = min_loc
        self.max_loc = max_loc
        self.penalty_factor = penalty_factor
        self.prize_required = prize_required
        self.check_solution = check_solution

    def _step(self, td: TensorDict) -> TensorDict:
        current_node = td["action"]

        # Get current coordinates, prize, and penalty
        cur_total_prize = td["cur_total_prize"] + gather_by_index(
            td["real_prize"], current_node
        )
        cur_total_penalty = td["cur_total_penalty"] + gather_by_index(
            td["penalty"], current_node
        )

        # Update visited
        visited = td["visited"].scatter(-1, current_node[..., None], 1)

        # Done and reward. Calculation is done outside hence set -inf
        done = (td["i"] > 0) & (current_node == 0)
        reward = torch.ones_like(cur_total_prize) * float("-inf")

        td_step = TensorDict(
            {
                "next": {
                    "locs": td["locs"],
                    "current_node": current_node,
                    "expected_prize": td["expected_prize"],
                    "real_prize": td["real_prize"],
                    "penalty": td["penalty"],
                    "cur_total_prize": cur_total_prize,
                    "cur_total_penalty": cur_total_penalty,
                    "visited": visited,
                    "prize_required": td["prize_required"],
                    "i": td["i"] + 1,
                    "reward": reward,
                    "done": done,
                },
            },
            batch_size=td.batch_size,
        )
        td_step["next"].set("action_mask", self.get_action_mask(td_step["next"]))
        return td_step

    def _reset(
        self, td: Optional[TensorDict] = None, batch_size: Optional[list] = None
    ) -> TensorDict:
        if batch_size is None:
            batch_size = self.batch_size if td is None else td["locs"].shape[:-2]
        if td is None or td.is_empty():
            td = self.generate_data(batch_size=batch_size)
        self.device = td.device

        locs = torch.cat([td["depot"][..., None, :], td["locs"]], dim=-2)
        expected_prize = td["deterministic_prize"]
        real_prize = (
            td["stochastic_prize"] if self.stochastic else td["deterministic_prize"]
        )
        penalty = td["penalty"]

        # Concatenate depots
        real_prize_with_depot = torch.cat(
            [torch.zeros_like(real_prize[..., :1]), real_prize], dim=-1
        )
        penalty_with_depot = F.pad(penalty, (1, 0), mode="constant", value=0)

        # Initialize the current node and  prize / penalty
        current_node = torch.zeros((*batch_size,), dtype=torch.int64, device=self.device)
        cur_total_prize = torch.zeros(*batch_size, device=self.device)
        cur_total_penalty = penalty.sum(-1)[
            :, None
        ]  # Sum penalties (all when nothing is visited), add step dim

        # Init the action mask (all nodes are available)
        visited = torch.zeros(
            (*batch_size, self.num_loc + 1), dtype=torch.bool, device=self.device
        )
        i = torch.zeros((*batch_size,), dtype=torch.int64, device=self.device)
        prize_required = torch.full(
            (*batch_size,), self.prize_required, device=self.device
        )

        td_reset = TensorDict(
            {
                "locs": locs,
                "current_node": current_node,
                "expected_prize": expected_prize,
                "real_prize": real_prize_with_depot,
                "penalty": penalty_with_depot,
                "cur_total_prize": cur_total_prize,
                "cur_total_penalty": cur_total_penalty,
                "visited": visited,
                "prize_required": prize_required,
                "i": i,
            },
            batch_size=batch_size,
        )
        td_reset.set("action_mask", self.get_action_mask(td_reset))
        return td_reset

    @staticmethod
    def get_action_mask(td: TensorDict) -> torch.Tensor:
        """Cannot visit depot if not yet collected 1 total prize and there are unvisited nodes"""
        mask = td["visited"] | td["visited"][..., 0:1]
        mask[..., 0] = (td["cur_total_prize"] < 1.0) & (
            td["visited"][..., 1:].int().sum(-1) < td["visited"][..., 1:].size(-1)
        )
        return ~(mask > 0)  # Invert mask, since 1 means feasible action

    def get_reward(self, td: TensorDict, actions: torch.Tensor) -> torch.Tensor:
        """Reward is `saved penalties - (total length + penalty)`"""

        # In case all tours directly return to depot, prevent further problems
        if actions.size(-1) == 1:
            assert (actions == 0).all(), "If all length 1 tours, they should be zero"
            return torch.zeros(actions.size(0), dtype=torch.float, device=actions.device)

        # Check that the solution is valid
        if self.check_solution:
            self.check_solution_validity(td, actions)

        # Gather locations in order of tour and get the length of tours
        locs_ordered = gather_by_index(td["locs"], actions)
        length = get_tour_length(locs_ordered)

        # Reward is saved penalties - (total length + penalty)
        saved_penalty = td["penalty"].gather(1, actions)
        return saved_penalty.sum(-1) - (length + td["penalty"][..., 1:].sum(-1))

    @staticmethod
    def check_solution_validity(td: TensorDict, actions: torch.Tensor):
        """Check that the solution is valid, i.e. contains all nodes once at most, and either prize constraint is met or all nodes are visited"""

        # Check that tours are valid, i.e. contain 0 to n -1
        sorted_actions = actions.data.sort(1)[0]

        # Make sure each node visited once at most (except for depot)
        assert (
            (sorted_actions[..., 1:] == 0)
            | (sorted_actions[..., 1:] > sorted_actions[..., :-1])
        ).all(), "Duplicates"

        prize = td["real_prize"][..., 1:]  # Remove depot
        prize_with_depot = torch.cat((torch.zeros_like(prize[:, :1]), prize), 1)
        p = prize_with_depot.gather(1, actions)

        # Either prize constraint should be satisfied or all prizes should be visited
        assert (
            (p.sum(-1) >= 1 - 1e-5)
            | (
                sorted_actions.size(-1) - (sorted_actions == 0).int().sum(-1)
                == (td["locs"].size(-2) - 1)
            )  # no depot
        ).all(), "Total prize does not satisfy min total prize"

    def generate_data(self, batch_size) -> TensorDict:
        # Batch size input check
        batch_size = [batch_size] if isinstance(batch_size, int) else batch_size

        depot = torch.rand((*batch_size, 2))
        locs = torch.rand((*batch_size, self.num_loc, 2))

        penalty_max = (
            MAX_LENGTHS[self.num_loc] * (self.penalty_factor) / float(self.num_loc)
        )
        penalty = torch.rand((*batch_size, self.num_loc)) * penalty_max

        # Take uniform prizes
        # Now expectation is 0.5 so expected total prize is n / 2, we want to force to visit approximately half of the nodes
        # so the constraint will be that total prize >= (n / 2) / 2 = n / 4
        # equivalently, we divide all prizes by n / 4 and the total prize should be >= 1
        deterministic_prize = (
            torch.rand((*batch_size, self.num_loc)) * 4 / float(self.num_loc)
        )

        # In the deterministic setting, the stochastic_prize is not used and the deterministic prize is known
        # In the stochastic setting, the deterministic prize is the expected prize and is known up front but the
        # stochastic prize is only revealed once the node is visited
        # Stochastic prize is between (0, 2 * expected_prize) such that E(stochastic prize) = E(deterministic_prize)
        stochastic_prize = (
            torch.rand((*batch_size, self.num_loc)) * deterministic_prize * 2
        )
        # In the deterministic setting, the stochastic_prize is not used and the deterministic prize is known
        # In the stochastic setting, the deterministic prize is the expected prize and is known up front but the
        # stochastic prize is only revealed once the node is visited
        # Stochastic prize is between (0, 2 * expected_prize) such that E(stochastic prize) = E(deterministic_prize)
        stochastic_prize = (
            torch.rand((*batch_size, self.num_loc)) * deterministic_prize * 2
        )

        return TensorDict(
            {
                "locs": locs,
                "depot": depot,
                "penalty": penalty,
                "deterministic_prize": deterministic_prize,
                "stochastic_prize": stochastic_prize,
            },
            batch_size=batch_size,
        )

    @property
    def stochastic(self):
        return self._stochastic

    @stochastic.setter
    def stochastic(self, state: bool):
        if state is True:
            log.warning(
                "Stochastic mode should not be used for PCTSP. Use SPCTSP instead."
            )

    def _make_spec(self, td_params: TensorDict):
        """Make the locs and action specs from the parameters."""
        self.observation_spec = CompositeSpec(
            locs=BoundedTensorSpec(
                minimum=self.min_loc,
                maximum=self.max_loc,
                shape=(self.num_loc, 2),
                dtype=torch.float32,
            ),
            current_node=UnboundedDiscreteTensorSpec(
                shape=(1),
                dtype=torch.int64,
            ),
            expected_prize=UnboundedContinuousTensorSpec(
                shape=(self.num_loc),
                dtype=torch.float32,
            ),
            real_prize=UnboundedContinuousTensorSpec(
                shape=(self.num_loc + 1),
                dtype=torch.float32,
            ),
            penalty=UnboundedContinuousTensorSpec(
                shape=(self.num_loc + 1),
                dtype=torch.float32,
            ),
            cur_total_prize=UnboundedContinuousTensorSpec(
                shape=(1),
                dtype=torch.float32,
            ),
            cur_total_penalty=UnboundedContinuousTensorSpec(
                shape=(1),
                dtype=torch.float32,
            ),
            visited=UnboundedDiscreteTensorSpec(
                shape=(self.num_loc + 1),
                dtype=torch.bool,
            ),
            prize_required=UnboundedContinuousTensorSpec(
                shape=(1),
                dtype=torch.float32,
            ),
            i=UnboundedDiscreteTensorSpec(
                shape=(1),
                dtype=torch.int64,
            ),
            action_mask=UnboundedDiscreteTensorSpec(
                shape=(self.num_loc),
                dtype=torch.bool,
            ),
            shape=(),
        )
        self.input_spec = self.observation_spec.clone()
        self.action_spec = BoundedTensorSpec(
            shape=(1,),
            dtype=torch.int64,
            minimum=0,
            maximum=self.num_loc,
        )
        self.reward_spec = UnboundedContinuousTensorSpec(shape=(1,))
        self.done_spec = UnboundedDiscreteTensorSpec(shape=(1,), dtype=torch.bool)

    @staticmethod
    def render(td, actions=None, ax=None):
        import matplotlib.pyplot as plt
        import numpy as np

        from matplotlib import colormaps

        # Create a plot of the nodes
        if ax is None:
            _, ax = plt.subplots()

        td = td.detach().cpu()

        # Actions
        if actions is None:
            actions = td.get("action", None)
        actions = actions.detach().cpu() if actions is not None else None

        # if batch_size greater than 0 , we need to select the first batch element
        if td.batch_size != torch.Size([]):
            td = td[0]
            actions = actions[0] if actions is not None else None

        # Variables
        depot = td["locs"][0, :]
        cities = td["locs"][1:, :]
        prizes = td["real_prize"][1:]
        penalties = td["penalty"][1:]
        normalized_prizes = (
            200 * (prizes - torch.min(prizes)) / (torch.max(prizes) - torch.min(prizes))
            + 10
        )
        normalized_penalties = (
            3
            * (penalties - torch.min(penalties))
            / (torch.max(penalties) - torch.min(penalties))
        )

        # Represent penalty with colormap and size of edges
        penalty_cmap = colormaps.get_cmap("BuPu")
        penalty_colors = penalty_cmap(normalized_penalties)

        # Plot depot and cities with prize (size of nodes) and penalties (size of borders)
        ax.scatter(
            depot[0],
            depot[1],
            marker="s",
            c="tab:green",
            edgecolors="black",
            zorder=1,
            s=100,
        )  # Plot depot as square
        ax.scatter(
            cities[:, 0],
            cities[:, 1],
            s=normalized_prizes,
            c=normalized_prizes,
            cmap="autumn_r",
            alpha=1,
            edgecolors=penalty_colors,
            linewidths=normalized_penalties,
        )  # Plot all cities with size and color indicating the prize

        # Gather locs in order of action if available
        if actions is None:
            print("No action in TensorDict, rendering unsorted locs")
        else:
            # Reorder the cities and their corresponding prizes based on actions
            tour = cities[actions - 1]  # subtract 1 to match Python's 0-indexing

            # Append the depot at the beginning and the end of the tour
            tour = np.vstack((depot, tour, depot))

            # Use quiver to plot the tour
            dx, dy = np.diff(tour[:, 0]), np.diff(tour[:, 1])
            ax.quiver(
                tour[:-1, 0],
                tour[:-1, 1],
                dx,
                dy,
                scale_units="xy",
                angles="xy",
                scale=1,
                zorder=2,
                color="black",
                width=0.0035,
            )

        # Setup limits and show
        ax.set_xlim(-0.05, 1.05)
        ax.set_ylim(-0.05, 1.05)
        plt.show()



# Base environment
# Main Environments
from rl4co.envs.atsp import ATSPEnv
from rl4co.envs.common.base import RL4COEnvBase
from rl4co.envs.cvrp import CVRPEnv
from rl4co.envs.dpp import DPPEnv
from rl4co.envs.ffsp import FFSPEnv
from rl4co.envs.mdpp import MDPPEnv
from rl4co.envs.mtsp import MTSPEnv
from rl4co.envs.op import OPEnv
from rl4co.envs.pctsp import PCTSPEnv
from rl4co.envs.pdp import PDPEnv
from rl4co.envs.sdvrp import SDVRPEnv
from rl4co.envs.spctsp import SPCTSPEnv
from rl4co.envs.tsp import TSPEnv

# Register environments
ENV_REGISTRY = {
    "atsp": ATSPEnv,
    "cvrp": CVRPEnv,
    "dpp": DPPEnv,
    "mdpp": MDPPEnv,
    "mtsp": MTSPEnv,
    "op": OPEnv,
    "pctsp": PCTSPEnv,
    "pdp": PDPEnv,
    "sdvrp": SDVRPEnv,
    "spctsp": SPCTSPEnv,
    "tsp": TSPEnv,
}


def get_env(env_name: str, *args, **kwargs) -> RL4COEnvBase:
    """Get environment by name.

    Args:
        env_name: Environment name
        *args: Positional arguments for environment
        **kwargs: Keyword arguments for environment

    Returns:
        Environment
    """
    env_cls = ENV_REGISTRY.get(env_name, None)
    if env_cls is None:
        raise ValueError(
            f"Unknown environment {env_name}. Available environments: {ENV_REGISTRY.keys()}"
        )
    return env_cls(*args, **kwargs)



from typing import Optional, Union

from rl4co.envs.common.base import RL4COEnvBase
from rl4co.models.rl import REINFORCE
from rl4co.models.rl.reinforce.baselines import REINFORCEBaseline
from rl4co.models.zoo.am.policy import AttentionModelPolicy


class EquityTransformer(REINFORCE):
    """Equity Transformer from Son et al., 2023.
    Reference: https://arxiv.org/abs/2306.02689

    Warning:
        This implementation is under development and subject to change.
    """

    def __init__(
        self,
        env: RL4COEnvBase,
        policy: Optional(AttentionModelPolicy) = None,
        baseline: Union[REINFORCEBaseline, str] = "rollout",
        policy_kwargs={},
        baseline_kwargs={},
        **kwargs,
    ):
        if policy is None:
            policy = AttentionModelPolicy(env.name, **policy_kwargs)

        super().__init__(env, policy, baseline, baseline_kwargs, **kwargs)



import hashlib
import logging
import os
import sys
import time
import urllib

from typing import Any, Optional
from urllib.parse import urlparse

import requests

from tqdm.auto import tqdm

from rl4co.utils.download.constants import USER_AGENT
from rl4co.utils.download.gdrive import get_url_filename_drive
from rl4co.utils.pylogger import get_pylogger

log = get_pylogger(__name__)


def download_url(
    url: str,
    root: str,
    filename: Optional[str] = None,
    session: requests.Session = requests.Session(),
    key: Optional[str] = None,
    proxy: str = None,
    md5: Optional[str] = None,
    max_redirect_hops: int = 3,
    verify: Optional[bool] = None,
    timeout: int = 60,
    retry_max: int = 500,
    sleep_max: int = 120,
    chunk_size: int = 1024,
    show_progress: bool = True,
    logging_level: int = logging.INFO,
) -> None:
    """Download a file from a url and place it in root. Supports robust downloads with resuming,
    connection error handling, proxies, authentication, and more.
    Args:
        url (str): URL to download file from
        root (str): Directory to place downloaded file in
        filename (str, optional): Name to save the file under. If None, use the basename of the URL
        session (requests.Session, optional): Session to use for HTTP requests.
        key (str, optional): authentication key in the form username:password
        proxy (str, optional): Proxy URL.
        md5 (str, optional): MD5 checksum of the download. If None, do not check
        max_redirect_hops (int, optional): Maximum number of redirects to follow
        verify (bool, optional): Whether to verify SSL certificates
        timeout (int, optional): Timeout for HTTP requests
        retry_max (int, optional): Maximum number of retries
        sleep_max (int, optional): Maximum number of seconds to sleep between retries
        chunk_size (int, optional): Number of bytes to read into memory at once
        show_progress (bool, optional): Whether to show a progress bar
        logging_level (int, optional): Logging level
    """
    # Set logging level
    log.setLevel(logging_level)

    # Expand redirect hops if needed
    url = _get_redirect_url(url, max_hops=max_redirect_hops)

    # Additional session configuration
    if key:
        session.auth = tuple(key.split(":", 2))
    if proxy:
        session.proxies = {"http": proxy, "https": proxy}

    # Try to get Google Drive url and if it exists, use it
    url_drive, filename_drive = get_url_filename_drive(url, session, verify)

    # Expand root and filename
    root = os.path.expanduser(root)
    if not filename and not url_drive:
        response = session.head(url, allow_redirects=True)
        content_disposition = response.headers.get("Content-Disposition")
        if content_disposition:
            filename = content_disposition.split(";")[-1]
        else:
            filename = os.path.basename(urlparse(response.url).path)
    elif url_drive:
        url = url_drive
        filename = filename_drive if not filename else filename

    fpath = os.path.join(root, filename)
    os.makedirs(root, exist_ok=True)

    # Check if file is already present locally
    if md5 is not None and check_integrity(fpath, md5):
        log.info("Using downloaded and verified file: " + fpath)
        return

    # Get file size
    try:
        downloaded = os.path.getsize(fpath)
    except FileNotFoundError:
        downloaded = 0
    size = int(session.get(url, stream=True).headers["Content-Length"])
    if downloaded == size:
        log.info("File %s already downloaded", filename)
        return
    elif downloaded > size:
        raise RuntimeError("File %s is corrupted" % filename)

    log.info("Downloading %s to %s (%s)", url, filename, format_bytes(size))

    mode = "ab"  # append to file if it exists
    sleep = 10
    tries = 0
    headers = {"Range": "bytes=%d-" % downloaded, "User-Agent": USER_AGENT}

    # Main loop: download the file in chunks with retries and resume on failure
    while tries < retry_max:
        r = robust_wrapper(session.get, retry_max, sleep_max)(
            url,
            stream=True,
            verify=verify,
            headers=headers,
            timeout=timeout,
        )
        try:
            r.raise_for_status()

            with tqdm(
                total=size,
                unit_scale=True,
                unit_divisor=1024,
                unit="B",
                initial=downloaded,
                disable=not show_progress,
            ) as pbar:
                with open(fpath, mode) as f:
                    for chunk in r.iter_content(chunk_size=chunk_size):
                        if chunk:
                            f.write(chunk)
                            downloaded += len(chunk)
                            pbar.update(len(chunk))

        except requests.exceptions.ConnectionError as e:
            log.error("Download interrupted: %s" % (e,))
        finally:
            r.close()

        if downloaded >= size:
            break

        log.error(
            "Download incomplete, downloaded %s / %s"
            % (format_bytes(downloaded), format_bytes(size))
        )
        log.warning("Sleeping %s seconds" % (sleep,))
        time.sleep(sleep)
        mode = "ab"

        downloaded = os.path.getsize(fpath)
        sleep *= 1.5
        if sleep > sleep_max:
            sleep = sleep_max
        headers = {"Range": "bytes=%d-" % downloaded, "User-Agent": USER_AGENT}
        tries += 1
        log.warning("Resuming from downloaded %s" % (format_bytes(downloaded),))

    if downloaded != size:
        raise Exception(
            "Download failed: downloaded %s / %s"
            % (format_bytes(downloaded), format_bytes(size))
        )

    if md5 is not None:
        if not check_integrity(fpath, md5):
            raise RuntimeError("File not found or corrupted.")
        else:
            log.info("File integrity verified!")


def robust_wrapper(call, retry_max=500, sleep_max=120):
    def retriable(code, reason):
        if code in [
            requests.codes.internal_server_error,
            requests.codes.bad_gateway,
            requests.codes.service_unavailable,
            requests.codes.gateway_timeout,
            requests.codes.too_many_requests,
            requests.codes.request_timeout,
        ]:
            return True

        return False

    def wrapped(*args, **kwargs):
        tries = 0
        while tries < retry_max:
            try:
                r = call(*args, **kwargs)
            except (
                requests.exceptions.ConnectionError,
                requests.exceptions.ReadTimeout,
            ) as e:
                r = None
                log.warning(
                    "Recovering from connection error [%s], attempts %s of %s",
                    e,
                    tries,
                    retry_max,
                )

            if r is not None:
                if not retriable(r.status_code, r.reason):
                    return r
                try:
                    log.warning(r.json()["reason"])
                except Exception:
                    pass
                log.warning(
                    "Recovering from HTTP error [%s %s], attempts %s of %s",
                    r.status_code,
                    r.reason,
                    tries,
                    retry_max,
                )

            tries += 1

            log.warning("Retrying in %s seconds", sleep_max)
            time.sleep(sleep_max)
            log.info("Retrying now...")

    return wrapped


def calculate_md5(fpath: str, chunk_size: int = 1024 * 1024) -> str:
    # Setting the `usedforsecurity` flag does not change anything about the functionality, but indicates that we are
    # not using the MD5 checksum for cryptography. This enables its usage in restricted environments like FIPS. Without
    # it torchvision.datasets is unusable in these environments since we perform a MD5 check everywhere.
    if sys.version_info >= (3, 9):
        md5 = hashlib.md5(usedforsecurity=False)
    else:
        md5 = hashlib.md5()
    with open(fpath, "rb") as f:
        for chunk in iter(lambda: f.read(chunk_size), b""):
            md5.update(chunk)
    return md5.hexdigest()


def _get_redirect_url(url: str, max_hops: int = 3) -> str:
    initial_url = url
    headers = {"Method": "HEAD", "User-Agent": USER_AGENT}

    for _ in range(max_hops + 1):
        with urllib.request.urlopen(
            urllib.request.Request(url, headers=headers)
        ) as response:
            if response.url == url or response.url is None:
                return url

            url = response.url
    else:
        raise RecursionError(
            f"Request to {initial_url} exceeded {max_hops} redirects. The last redirect points to {url}."
        )


def check_md5(fpath: str, md5: str, **kwargs: Any) -> bool:
    return md5 == calculate_md5(fpath, **kwargs)


def check_integrity(fpath: str, md5: Optional[str] = None) -> bool:
    if not os.path.isfile(fpath):
        return False
    if md5 is None:
        return True
    return check_md5(fpath, md5)


def format_bytes(num_bytes):
    units = ["", "K", "M", "G", "T", "P"]
    unit_index = 0
    bytes_remaining = num_bytes
    while bytes_remaining >= 1024:
        bytes_remaining /= 1024.0
        unit_index += 1
    return f"{int(bytes_remaining * 10 + 0.5) / 10.0}{units[unit_index]}"



from typing import Union

from tensordict import TensorDict
from torch import Tensor, nn

from rl4co.models.nn.env_embeddings import env_init_embedding
from rl4co.models.nn.graph.attnnet import GraphAttentionNetwork


class CriticNetwork(nn.Module):
    """We make the critic network compatible with any problem by using encoder for any environment
    Refactored from Kool et al. (2019) which only worked for TSP. In our case, we make it
    compatible with any problem by using the environment init embedding.

    Args:
        env_name: environment name to solve
        encoder: Encoder to use for the critic
        embedding_dim: Dimension of the embeddings
        hidden_dim: Hidden dimension for the feed-forward network
        num_layers: Number of layers for the encoder
        num_heads: Number of heads for the attention
        normalization: Normalization to use for the attention
        force_flash_attn: Whether to force the use of flash attention. If True, cast to fp16
    """

    def __init__(
        self,
        env_name: str = None,
        encoder: nn.Module = None,
        embedding_dim: int = 128,
        hidden_dim: int = 512,
        num_layers: int = 3,
        num_heads: int = 8,
        normalization: str = "batch",
        force_flash_attn: bool = False,
        **unused_kwargs,
    ):
        super(CriticNetwork, self).__init__()

        if env_name is None:
            self.init_embedding = nn.Identity()
        else:
            self.init_embedding = env_init_embedding(
                env_name, {"embedding_dim": embedding_dim}
            )

        self.encoder = (
            GraphAttentionNetwork(
                num_heads=num_heads,
                embedding_dim=embedding_dim,
                num_layers=num_layers,
                normalization=normalization,
                feed_forward_hidden=hidden_dim,
                force_flash_attn=force_flash_attn,
            )
            if encoder is None
            else encoder
        )

        self.value_head = nn.Sequential(
            nn.Linear(embedding_dim, hidden_dim), nn.ReLU(), nn.Linear(hidden_dim, 1)
        )

    def forward(self, x: Union[Tensor, TensorDict]) -> Tensor:
        """Forward pass of the critic network: encode the imput in embedding space and return the value

        Args:
            x: Input containing the environment state. Can be a Tensor or a TensorDict

        Returns:
            Value of the input state
        """

        # Initial embedding of x. This is the identity function if env_name is None.
        x = self.init_embedding(x)
        x = self.encoder(x)
        return self.value_head(x).mean(1)



from pathlib import Path
from typing import Sequence

import rich
import rich.syntax
import rich.tree

from hydra.core.hydra_config import HydraConfig
from lightning.pytorch.utilities.rank_zero import rank_zero_only
from omegaconf import DictConfig, OmegaConf, open_dict
from rich.prompt import Prompt

from rl4co.utils.utils import pylogger

log = pylogger.get_pylogger(__name__)


@rank_zero_only
def print_config_tree(
    cfg: DictConfig,
    print_order: Sequence[str] = (
        # "data", # note: data is dealt with in model
        "model",
        "callbacks",
        "logger",
        "trainer",
        "paths",
        "extras",
    ),
    resolve: bool = True,
    save_to_file: bool = False,
) -> None:
    """Prints content of DictConfig using Rich library and its tree structure.
    Args:
        cfg (DictConfig): Configuration composed by Hydra.
        print_order (Sequence[str], optional): Determines in what order config components are printed.
        resolve (bool, optional): Whether to resolve reference fields of DictConfig.
        save_to_file (bool, optional): Whether to export config to the hydra output folder.
    """

    style = "dim"
    tree = rich.tree.Tree("CONFIG", style=style, guide_style=style)

    queue = []

    # add fields from `print_order` to queue
    for field in print_order:
        queue.append(field) if field in cfg else log.warning(
            f"Field '{field}' not found in config. Skipping '{field}' config printing..."
        )

    # add all the other fields to queue (not specified in `print_order`)
    for field in cfg:
        if field not in queue:
            queue.append(field)

    # generate config tree from queue
    for field in queue:
        branch = tree.add(field, style=style, guide_style=style)

        config_group = cfg[field]
        if isinstance(config_group, DictConfig):
            branch_content = OmegaConf.to_yaml(config_group, resolve=resolve)
        else:
            branch_content = str(config_group)

        branch.add(rich.syntax.Syntax(branch_content, "yaml"))

    # print config tree
    rich.print(tree)

    # save config tree to file
    if save_to_file:
        with open(Path(cfg.paths.output_dir, "config_tree.log"), "w") as file:
            rich.print(tree, file=file)


@rank_zero_only
def enforce_tags(cfg: DictConfig, save_to_file: bool = False) -> None:
    """Prompts user to input tags from command line if no tags are provided in config."""

    if not cfg.get("tags"):
        if "id" in HydraConfig().cfg.hydra.job:
            raise ValueError("Specify tags before launching a multirun!")

        log.warning("No tags provided in config. Prompting user to input tags...")
        tags = Prompt.ask("Enter a list of comma separated tags", default="dev")
        tags = [t.strip() for t in tags.split(",") if t != ""]

        with open_dict(cfg):
            cfg.tags = tags

        log.info(f"Tags: {cfg.tags}")

    if save_to_file:
        with open(Path(cfg.paths.output_dir, "tags.log"), "w") as file:
            rich.print(cfg.tags, file=file)



import torch.nn as nn

from rl4co.utils.pylogger import get_pylogger

log = get_pylogger(__name__)


def env_dynamic_embedding(env_name: str, config: dict) -> nn.Module:
    """Get environment dynamic embedding. The dynamic embedding is used to modify query, key and value vectors of the attention mechanism
    based on the current state of the environment (which is changing during the rollout).
    Consists of a linear layer that projects the node features to the embedding space.

    Args:
        env: Environment or its name.
        config: A dictionary of configuration options for the environment.
    """
    embedding_registry = {
        "tsp": StaticEmbedding,
        "atsp": StaticEmbedding,
        "cvrp": StaticEmbedding,
        "sdvrp": SDVRPDynamicEmbedding,
        "pctsp": StaticEmbedding,
        "spctsp": StaticEmbedding,
        "op": StaticEmbedding,
        "dpp": StaticEmbedding,
        "mdpp": StaticEmbedding,
        "pdp": StaticEmbedding,
        "mtsp": StaticEmbedding,
    }

    if env_name not in embedding_registry:
        log.warning(
            f"Unknown environment name '{env_name}'. Available dynamic embeddings: {embedding_registry.keys()}. Defaulting to StaticEmbedding."
        )
    return embedding_registry.get(env_name, StaticEmbedding)(**config)


class StaticEmbedding(nn.Module):
    """Static embedding for general problems.
    This is used for problems that do not have any dynamic information, except for the
    information regarding the current action (e.g. the current node in TSP). See context embedding for more details.
    """

    def __init__(self, *args, **kwargs):
        super(StaticEmbedding, self).__init__()

    def forward(self, td):
        return 0, 0, 0


class SDVRPDynamicEmbedding(nn.Module):
    """Dynamic embedding for the Split Delivery Vehicle Routing Problem (SDVRP).
    Embed the following node features to the embedding space:
        - demand_with_depot: demand of the customers and the depot
    The demand with depot is used to modify the query, key and value vectors of the attention mechanism
    based on the current state of the environment (which is changing during the rollout).
    """

    def __init__(self, embedding_dim):
        super(SDVRPDynamicEmbedding, self).__init__()
        self.projection = nn.Linear(1, 3 * embedding_dim, bias=False)

    def forward(self, td):
        demands_with_depot = td["demand_with_depot"][..., None].clone()
        demands_with_depot[..., 0, :] = 0
        glimpse_key_dynamic, glimpse_val_dynamic, logit_key_dynamic = self.projection(
            demands_with_depot
        ).chunk(3, dim=-1)
        return glimpse_key_dynamic, glimpse_val_dynamic, logit_key_dynamic



from typing import List, Optional, Tuple

import hydra
import lightning as L
import pyrootutils
import torch

from lightning import Callback, LightningModule
from lightning.pytorch.loggers import Logger
from omegaconf import DictConfig

pyrootutils.setup_root(__file__, indicator=".gitignore", pythonpath=True)

from rl4co import utils
from rl4co.utils import RL4COTrainer

log = utils.get_pylogger(__name__)


@utils.task_wrapper
def run(cfg: DictConfig) -> Tuple[dict, dict]:
    """Trains the model. Can additionally evaluate on a testset, using best weights obtained during
    training.
    This method is wrapped in optional @task_wrapper decorator, that controls the behavior during
    failure. Useful for multiruns, saving info about the crash, etc.

    Args:
        cfg (DictConfig): Configuration composed by Hydra.
    Returns:
        Tuple[dict, dict]: Dict with metrics and dict with all instantiated objects.
    """

    # set seed for random number generators in pytorch, numpy and python.random
    if cfg.get("seed"):
        L.seed_everything(cfg.seed, workers=True)

    # We instantiate the environment separately and then pass it to the model
    log.info(f"Instantiating environment <{cfg.env._target_}>")
    env = hydra.utils.instantiate(cfg.env)

    # Note that the RL environment is instantiated inside the model
    log.info(f"Instantiating model <{cfg.model._target_}>")
    model: LightningModule = hydra.utils.instantiate(cfg.model, env)

    log.info("Instantiating callbacks...")
    callbacks: List[Callback] = utils.instantiate_callbacks(cfg.get("callbacks"))

    log.info("Instantiating loggers...")
    logger: List[Logger] = utils.instantiate_loggers(cfg.get("logger"))

    log.info("Instantiating trainer...")
    trainer: RL4COTrainer = hydra.utils.instantiate(
        cfg.trainer,
        callbacks=callbacks,
        logger=logger,
    )

    object_dict = {
        "cfg": cfg,
        "model": model,
        "callbacks": callbacks,
        "logger": logger,
        "trainer": trainer,
    }

    if logger:
        log.info("Logging hyperparameters!")
        utils.log_hyperparameters(object_dict)

    if cfg.get("compile", False):
        log.info("Compiling model!")
        model = torch.compile(model)

    if cfg.get("train"):
        log.info("Starting training!")
        trainer.fit(model=model, ckpt_path=cfg.get("ckpt_path"))

        train_metrics = trainer.callback_metrics

    if cfg.get("test"):
        log.info("Starting testing!")
        ckpt_path = trainer.checkpoint_callback.best_model_path
        if ckpt_path == "":
            log.warning("Best ckpt not found! Using current weights for testing...")
            ckpt_path = None
        trainer.test(model=model, ckpt_path=ckpt_path)
        log.info(f"Best ckpt path: {ckpt_path}")

    test_metrics = trainer.callback_metrics

    # merge train and test metrics
    metric_dict = {**train_metrics, **test_metrics}

    return metric_dict, object_dict


@hydra.main(version_base="1.3", config_path="../../configs", config_name="main.yaml")
# @hydra.main(version_base="1.3", config_path="configs", config_name="experiment/tsp/am-ppo.yaml")
def train(cfg: DictConfig) -> Optional[float]:
    # apply extra utilities
    # (e.g. ask for tags if none are provided in cfg, print cfg tree, etc.)
    utils.extras(cfg)

    # train the model
    metric_dict, _ = run(cfg)

    # safely retrieve metric value for hydra-based hyperparameter optimization
    metric_value = utils.get_metric_value(
        metric_dict=metric_dict, metric_name=cfg.get("optimized_metric")
    )

    # return optimized metric
    return metric_value


if __name__ == "__main__":
    train()



import math

import torch
import torch.nn as nn


class Encoder(nn.Module):
    """Maps a graph represented as an input sequence
    to a hidden vector"""

    def __init__(self, input_dim, hidden_dim):
        super(Encoder, self).__init__()
        self.hidden_dim = hidden_dim
        self.lstm = nn.LSTM(input_dim, hidden_dim)
        self.init_hx, self.init_cx = self.init_hidden(hidden_dim)

    def forward(self, x, hidden):
        output, hidden = self.lstm(x, hidden)
        return output, hidden

    def init_hidden(self, hidden_dim):
        """Trainable initial hidden state"""
        std = 1.0 / math.sqrt(hidden_dim)
        enc_init_hx = nn.Parameter(torch.FloatTensor(hidden_dim))
        enc_init_hx.data.uniform_(-std, std)

        enc_init_cx = nn.Parameter(torch.FloatTensor(hidden_dim))
        enc_init_cx.data.uniform_(-std, std)
        return enc_init_hx, enc_init_cx



from typing import Optional

import torch

from tensordict.tensordict import TensorDict
from torchrl.data import (
    BoundedTensorSpec,
    CompositeSpec,
    UnboundedContinuousTensorSpec,
    UnboundedDiscreteTensorSpec,
)

from rl4co.envs.common.base import RL4COEnvBase
from rl4co.utils.ops import gather_by_index
from rl4co.utils.pylogger import get_pylogger

log = get_pylogger(__name__)


class MPDPEnv(RL4COEnvBase):
    """Multi-agent Pickup and Delivery Problem environment.
    The goal is to pick up and deliver all the packages while satisfying the precedence constraints.
    When an agent goes back to the depot, a new agent is spawned. In the min-max version, the goal is to minimize the
    maximum tour length among all agents.
    The reward is the -infinite unless the agent visits all the cities.
    In that case, the reward is (-)length of the path: maximizing the reward is equivalent to minimizing the path length.

    Args:
        num_loc: number of locations (cities) in the TSP
        min_loc: minimum location coordinate. Used for data generation
        max_loc: maximum location coordinate. Used for data generation
        min_num_agents: minimum number of agents. Used for data generation
        max_num_agents: maximum number of agents. Used for data generation
        objective: objective to optimize. Either 'minmax' or 'minsum'
        check_solution: whether to check the validity of the solution
        td_params: parameters of the environment
    """

    name = "mpdp"

    def __init__(
        self,
        num_loc: int = 20,
        min_loc: float = 0,
        max_loc: float = 1,
        min_num_agents: int = 2,
        max_num_agents: int = 10,
        objective: str = "minmax",
        check_solution: bool = False,
        td_params: TensorDict = None,
        **kwargs,
    ):
        super().__init__(**kwargs)
        self.num_loc = num_loc
        self.min_loc = min_loc
        self.max_loc = max_loc
        self.min_num_agents = min_num_agents
        self.max_num_agents = max_num_agents
        self.objective = objective
        self.check_solution = check_solution
        self._make_spec(td_params)

    def _step(self, td: TensorDict) -> TensorDict:
        selected = td["action"][:, None]  # Add dimension for step

        agent_num = td["lengths"].size(1)
        n_loc = td["to_delivery"].size(-1) - agent_num - 1

        new_to_delivery = (selected + n_loc // 2) % (
            n_loc + agent_num + 1
        )  # the pair node of selected node

        is_request = (selected > agent_num) & (selected <= agent_num + n_loc // 2)
        td["left_request"][is_request] -= 1
        depot_distance = td["depot_distance"].scatter(-1, selected, 0)

        add_pd = td["add_pd_distance"][is_request.squeeze(-1), :].gather(
            -1, selected[is_request.squeeze(-1), :] - agent_num - 1
        )
        td["longest_lengths"][is_request.squeeze(-1), :].scatter_add_(
            -1, td["count_depot"][is_request.squeeze(-1), :], add_pd
        )
        td["add_pd_distance"][is_request.squeeze(-1), :].scatter_(
            -1, selected[is_request.squeeze(-1), :] - agent_num - 1, 0
        )
        remain_sum_paired_distance = td["add_pd_distance"].sum(-1, keepdim=True)
        remain_pickup_max_distance = depot_distance[:, : agent_num + 1 + n_loc // 2].max(
            dim=-1, keepdim=True
        )[0]
        remain_delivery_max_distance = depot_distance[
            :, agent_num + 1 + n_loc // 2 :
        ].max(dim=-1, keepdim=True)[0]

        # Calculate makespan
        cur_coord = gather_by_index(td["locs"], selected)
        path_lengths = (cur_coord - td["cur_coord"]).norm(p=2, dim=-1)

        td["lengths"].scatter_add_(-1, td["count_depot"], path_lengths.unsqueeze(-1))

        # If visit depot then plus one to count_depot\
        td["count_depot"][
            (selected == td["agent_idx"]) & (td["agent_idx"] < agent_num)
        ] += 1  # torch.ones(td["count_depot"][(selected == 0) & (td["agent_idx"] < agent_num)].shape, dtype=torch.int64, device=td["count_depot"].device)

        # `agent_idx` is added by 1 if the current agent comes back to depot
        agent_idx = (td["count_depot"] + 1) * torch.ones(
            selected.size(0), 1, dtype=torch.long, device=td["count_depot"].device
        )
        visited = td["visited"].scatter(-1, selected.unsqueeze(-1), 1)
        to_delivery = td["to_delivery"].scatter(-1, new_to_delivery[:, :, None], 1)

        # Get done and reward
        done = visited.all(dim=-1, keepdim=True).squeeze(-1)
        reward = torch.ones_like(done) * float(
            "-inf"
        )  # reward calculated via `get_reward` for now

        td_step = TensorDict(
            {
                "next": {
                    "locs": td["locs"],
                    "visited": visited,
                    "lengths": td["lengths"],
                    "count_depot": td["count_depot"],
                    "agent_idx": agent_idx,
                    "cur_coord": cur_coord,
                    "to_delivery": to_delivery,
                    "left_request": td["left_request"],
                    "depot_distance": depot_distance,
                    "remain_sum_paired_distance": remain_sum_paired_distance,
                    "remain_pickup_max_distance": remain_pickup_max_distance,
                    "remain_delivery_max_distance": remain_delivery_max_distance,
                    "add_pd_distance": td["add_pd_distance"],
                    "longest_lengths": td["longest_lengths"],
                    "i": td["i"] + 1,
                    "done": done,
                    "reward": reward,
                }
            },
            td.shape,
        )
        td_step["next"].set("action_mask", self.get_action_mask(td_step["next"]))
        return td_step

    def _reset(
        self,
        td: Optional[TensorDict] = None,
        batch_size: Optional[list] = None,
        agent_num: Optional[int] = None,  # NOTE hardcoded from ET
    ) -> TensorDict:
        if batch_size is None:
            batch_size = self.batch_size if td is None else td["locs"].shape[:-2]

        if td is None or td.is_empty():
            td = self.generate_data(batch_size=batch_size)

        self.device = td.device

        # NOTE: this is a hack to get the agent_num
        # agent_num = td["agent_num"][0].item() if agent_num is None else agent_num
        # agent_num = agent_num if agent_num is not None else td["agent_num"][0].item()

        depot = td["depot"]
        depot = depot.repeat(1, agent_num + 1, 1)
        loc = td["locs"]
        left_request = loc.size(1) // 2
        whole_instance = torch.cat((depot, loc), dim=1)

        # Distance from all nodes between each other
        distance = torch.cdist(whole_instance, whole_instance, p=2)
        index = torch.arange(left_request, 2 * left_request, device=depot.device)[
            None, :, None
        ]
        index = index.repeat(distance.shape[0], 1, 1)
        add_pd_distance = distance[
            :, agent_num + 1 : agent_num + 1 + left_request, agent_num + 1 :
        ].gather(-1, index)
        add_pd_distance = add_pd_distance.squeeze(-1)

        remain_pickup_max_distance = distance[:, 0, : agent_num + 1 + left_request].max(
            dim=-1, keepdim=True
        )[0]
        remain_delivery_max_distance = distance[:, 0, agent_num + 1 + left_request :].max(
            dim=-1, keepdim=True
        )[0]
        remain_sum_paired_distance = add_pd_distance.sum(dim=-1, keepdim=True)

        # Distance from depot to all nodes
        # Delivery nodes should consider the sum of distance from depot to paired pickup nodes and pickup nodes to delivery nodes
        distance[:, 0, agent_num + 1 : agent_num + 1 + left_request] = (
            distance[:, 0, agent_num + 1 : agent_num + 1 + left_request]
            + distance[:, 0, agent_num + 1 + left_request :]
        )

        # Distance from depot to all nodes
        depot_distance = distance[:, 0, :]
        depot_distance[:, agent_num + 1 : agent_num + 1 + left_request] = depot_distance[
            :, agent_num + 1 : agent_num + 1 + left_request
        ]  # + add_pd_distance

        batch_size, n_loc, _ = loc.size()
        to_delivery = torch.cat(
            [
                torch.ones(
                    batch_size,
                    1,
                    n_loc // 2 + agent_num + 1,
                    dtype=torch.uint8,
                    device=loc.device,
                ),
                torch.zeros(
                    batch_size, 1, n_loc // 2, dtype=torch.uint8, device=loc.device
                ),
            ],
            dim=-1,
        )

        # Create reset TensorDict
        td_reset = TensorDict(
            {
                "locs": torch.cat((depot, loc), -2),
                "visited": torch.zeros(
                    batch_size,
                    1,
                    n_loc + agent_num + 1,
                    dtype=torch.uint8,
                    device=loc.device,
                ),
                "lengths": torch.zeros(batch_size, agent_num, device=loc.device),
                "longest_lengths": torch.zeros(batch_size, agent_num, device=loc.device),
                "cur_coord": td["depot"]
                if len(td["depot"].shape) == 2
                else td["depot"].squeeze(1),
                "i": torch.zeros(
                    batch_size, dtype=torch.int64, device=loc.device
                ),  # Vector with length num_steps
                "to_delivery": to_delivery,
                "count_depot": torch.zeros(
                    batch_size, 1, dtype=torch.int64, device=loc.device
                ),
                "agent_idx": torch.ones(
                    batch_size, 1, dtype=torch.long, device=loc.device
                ),
                "left_request": left_request
                * torch.ones(batch_size, 1, dtype=torch.long, device=loc.device),
                "remain_pickup_max_distance": remain_pickup_max_distance,
                "remain_delivery_max_distance": remain_delivery_max_distance,
                "depot_distance": depot_distance,
                "remain_sum_paired_distance": remain_sum_paired_distance,
                "add_pd_distance": add_pd_distance,
            },
            batch_size=batch_size,
        )
        td_reset.set("action_mask", self.get_action_mask(td_reset))
        return td_reset

    @staticmethod
    def get_action_mask(td: TensorDict) -> torch.Tensor:
        """Get the action mask for the current state."""

        visited_loc = td["visited"].clone()

        agent_num = td["lengths"].size(1)
        n_loc = visited_loc.size(-1) - agent_num - 1  # num of customers
        batch_size = visited_loc.size(0)
        agent_idx = td["agent_idx"][:, None, :]
        mask_loc = visited_loc.to(td["to_delivery"].device) | (1 - td["to_delivery"])

        # depot
        if td["i"][0].item() != 0:
            mask_loc[:, :, : agent_num + 1] = 1

            # if deliver nodes which is assigned agent is complete, then agent can go to depot
            no_item_to_delivery = (
                visited_loc[:, :, n_loc // 2 + agent_num + 1 :]
                == td["to_delivery"][:, :, n_loc // 2 + agent_num + 1 :]
            ).all(dim=-1)
            mask_loc[no_item_to_delivery.squeeze(-1), :, :] = mask_loc[
                no_item_to_delivery.squeeze(-1), :, :
            ].scatter_(-1, agent_idx[no_item_to_delivery.squeeze(-1), :, :], 0)

            condition = (td["count_depot"] == agent_num - 1) & (
                (visited_loc[:, :, agent_num + 1 :] == 0).sum(dim=-1) != 0
            )

            mask_loc[..., agent_num][condition] = 1

        else:
            return (
                torch.cat(
                    [
                        torch.zeros(
                            batch_size, 1, 1, dtype=torch.uint8, device=mask_loc.device
                        ),
                        torch.ones(
                            batch_size,
                            1,
                            n_loc + agent_num,
                            dtype=torch.uint8,
                            device=mask_loc.device,
                        ),
                    ],
                    dim=-1,
                )
                > 0
            )
        action_mask = mask_loc == 0  # action_mask gets feasible actions
        return action_mask

    def get_reward(self, td: TensorDict, actions: TensorDict) -> TensorDict:
        # Check that the solution is valid
        if self.check_solution:
            self.check_solution_validity(td, actions)

        # Calculate the reward (negative tour length)
        if self.objective == "minmax":
            return -td["lengths"].max(dim=-1, keepdim=True)[0].squeeze(-1)
        elif self.objective == "minsum":
            return -td["lengths"].sum(dim=-1, keepdim=True).squeeze(-1)
        else:
            raise ValueError(f"Unknown objective {self.objective}")

    @staticmethod
    def check_solution_validity(td: TensorDict, actions: torch.Tensor):
        assert True, "Not implemented"

    def generate_data(self, batch_size) -> TensorDict:
        # Batch size input check
        batch_size = [batch_size] if isinstance(batch_size, int) else batch_size

        # Initialize the locations (including the depot which is always the first node)
        locs_with_depot = (
            torch.FloatTensor(*batch_size, self.num_loc + 1, 2)
            .uniform_(self.min_loc, self.max_loc)
            .to(self.device)
        )

        return TensorDict(
            {
                "locs": locs_with_depot[..., 1:, :],
                "depot": locs_with_depot[..., 0, :],
            },
            batch_size=batch_size,
        )

    def _make_spec(self, td_params: TensorDict):
        """Make the observation and action specs from the parameters."""
        max_nodes = self.num_loc + self.max_num_agents + 1
        self.observation_spec = CompositeSpec(
            locs=BoundedTensorSpec(
                minimum=self.min_loc,
                maximum=self.max_loc,
                shape=(max_nodes, 2),
                dtype=torch.float32,
            ),
            current_node=UnboundedDiscreteTensorSpec(
                shape=(1),
                dtype=torch.int64,
            ),
            action_mask=UnboundedDiscreteTensorSpec(
                shape=(max_nodes, 1),
                dtype=torch.bool,
            ),
            visited=UnboundedDiscreteTensorSpec(
                shape=(1, max_nodes),
                dtype=torch.bool,
            ),
            lengths=UnboundedContinuousTensorSpec(
                shape=(self.max_num_agents,),
                dtype=torch.float32,
            ),
            longest_lengths=UnboundedContinuousTensorSpec(
                shape=(self.max_num_agents,),
                dtype=torch.float32,
            ),
            cur_coord=BoundedTensorSpec(
                minimum=self.min_loc,
                maximum=self.max_loc,
                shape=(2,),
                dtype=torch.float32,
            ),
            to_delivery=UnboundedDiscreteTensorSpec(
                shape=(max_nodes, 1),
                dtype=torch.bool,
            ),
            count_depot=UnboundedDiscreteTensorSpec(
                shape=(1,),
                dtype=torch.int64,
            ),
            agent_idx=UnboundedDiscreteTensorSpec(
                shape=(1,),
                dtype=torch.int64,
            ),
            left_request=UnboundedDiscreteTensorSpec(
                shape=(1,),
                dtype=torch.int64,
            ),
            remain_pickup_max_distance=UnboundedContinuousTensorSpec(
                shape=(1,),
                dtype=torch.float32,
            ),
            remain_delivery_max_distance=UnboundedContinuousTensorSpec(
                shape=(1,),
                dtype=torch.float32,
            ),
            depot_distance=UnboundedContinuousTensorSpec(
                shape=(max_nodes,),
                dtype=torch.float32,
            ),
            remain_sum_paired_distance=UnboundedContinuousTensorSpec(
                shape=(1,),
                dtype=torch.float32,
            ),
            add_pd_distance=UnboundedContinuousTensorSpec(
                shape=(max_nodes,),
                dtype=torch.float32,
            ),
            ## NOTE: we should have a vectorized implementation for agent_num
            # agent_num=UnboundedDiscreteTensorSpec(
            #     shape=(1,),
            #     dtype=torch.int64,
            # ),
            i=UnboundedDiscreteTensorSpec(
                shape=(1,),
                dtype=torch.int64,
            ),
        )
        self.input_spec = self.observation_spec.clone()
        self.action_spec = BoundedTensorSpec(
            shape=(1,),
            dtype=torch.int64,
            minimum=0,
            maximum=max_nodes,
        )
        self.reward_spec = UnboundedContinuousTensorSpec(shape=(1,))
        self.done_spec = UnboundedDiscreteTensorSpec(shape=(1,), dtype=torch.bool)

    @staticmethod
    def render(td: TensorDict, actions=None, ax=None):
        # TODO: color switch with new agents; add pickup and delivery nodes as in `PDPEnv.render`

        import matplotlib.pyplot as plt
        import numpy as np

        from matplotlib import cm, colormaps

        num_routine = (actions == 0).sum().item() + 2
        base = colormaps["nipy_spectral"]
        color_list = base(np.linspace(0, 1, num_routine))
        cmap_name = base.name + str(num_routine)
        out = base.from_list(cmap_name, color_list, num_routine)

        if ax is None:
            # Create a plot of the nodes
            _, ax = plt.subplots()

        td = td.detach().cpu()

        if actions is None:
            actions = td.get("action", None)

        # if batch_size greater than 0 , we need to select the first batch element
        if td.batch_size != torch.Size([]):
            td = td[0]
            actions = actions[0]

        locs = td["locs"]

        # add the depot at the first action and the end action
        actions = torch.cat([torch.tensor([0]), actions, torch.tensor([0])])

        # gather locs in order of action if available
        if actions is None:
            log.warning("No action in TensorDict, rendering unsorted locs")
        else:
            locs = locs

        # Cat the first node to the end to complete the tour
        x, y = locs[:, 0], locs[:, 1]

        # plot depot
        ax.scatter(
            locs[0, 0],
            locs[0, 1],
            edgecolors=cm.Set2(2),
            facecolors="none",
            s=100,
            linewidths=2,
            marker="s",
            alpha=1,
        )

        # plot visited nodes
        ax.scatter(
            x[1:],
            y[1:],
            edgecolors=cm.Set2(0),
            facecolors="none",
            s=50,
            linewidths=2,
            marker="o",
            alpha=1,
        )

        # text depot
        ax.text(
            locs[0, 0],
            locs[0, 1] - 0.025,
            "Depot",
            horizontalalignment="center",
            verticalalignment="top",
            fontsize=10,
            color=cm.Set2(2),
        )

        # plot actions
        color_idx = 0
        for action_idx in range(len(actions) - 1):
            if actions[action_idx] == 0:
                color_idx += 1
            from_loc = locs[actions[action_idx]]
            to_loc = locs[actions[action_idx + 1]]
            ax.plot(
                [from_loc[0], to_loc[0]],
                [from_loc[1], to_loc[1]],
                color=out(color_idx),
                lw=1,
            )
            ax.annotate(
                "",
                xy=(to_loc[0], to_loc[1]),
                xytext=(from_loc[0], from_loc[1]),
                arrowprops=dict(arrowstyle="-|>", color=out(color_idx)),
                size=15,
                annotation_clip=False,
            )

        # Setup limits and show
        ax.set_xlim(-0.05, 1.05)
        ax.set_ylim(-0.05, 1.05)
        plt.show()



from rl4co.utils.instantiators import instantiate_callbacks, instantiate_loggers
from rl4co.utils.pylogger import get_pylogger
from rl4co.utils.rich_utils import enforce_tags, print_config_tree
from rl4co.utils.trainer import RL4COTrainer
from rl4co.utils.utils import extras, get_metric_value, log_hyperparameters, task_wrapper



from rl4co.models.zoo.am import AttentionModel, AttentionModelPolicy
from rl4co.models.zoo.common.autoregressive import AutoregressivePolicy
from rl4co.models.zoo.ppo import PPOModel, PPOPolicy
from rl4co.models.zoo.ptrnet import PointerNetwork, PointerNetworkPolicy
from rl4co.models.zoo.symnco import SymNCO, SymNCOPolicy
from rl4co.models.zoo.ham import HeterogeneousAttentionModel, HeterogeneousAttentionModelPolicy
from rl4co.models.zoo.mdam import MDAM, MDAMPolicy


from typing import Tuple, Union

from tensordict import TensorDict
from torch import Tensor

from rl4co.envs import RL4COEnvBase
from rl4co.models.zoo.common.autoregressive import AutoregressivePolicy
from rl4co.models.zoo.ppo.decoder import PPODecoder


class PPOPolicy(AutoregressivePolicy):
    """PPO Policy based on Kool et al. (2019): https://arxiv.org/abs/1803.08475.
    PPOPolicy supports 'evaluate_action' method to evaluate the action probability

    Args:
        env_name: Name of the environment used to initialize embeddings
        embedding_dim: Dimension of the node embeddings
        num_encoder_layers: Number of layers in the encoder
        num_heads: Number of heads in the attention layers
        normalization: Normalization type in the attention layers
        **kwargs: keyword arguments passed to the `AutoregressivePolicy`
    """

    def __init__(
        self,
        env_name: str,
        embedding_dim: int = 128,
        num_encoder_layers: int = 3,
        num_heads: int = 8,
        normalization: str = "batch",
        **kwargs,
    ):
        super(PPOPolicy, self).__init__(
            env_name=env_name,
            decoder=PPODecoder(
                env_name=env_name,
                embedding_dim=embedding_dim,
                num_heads=num_heads,
                **kwargs,
            ),  # override decoder with PPODecoder to support 'evaluate_action"
            embedding_dim=embedding_dim,
            num_encoder_layers=num_encoder_layers,
            num_heads=num_heads,
            normalization=normalization,
            **kwargs,
        )

    def evaluate_action(
        self,
        td: TensorDict,
        action: Tensor,
        env: Union[str, RL4COEnvBase] = None,
    ) -> Tuple[Tensor, Tensor]:
        embeddings, _ = self.encoder(td)
        ll, entropy = self.decoder.evaluate_action(td, embeddings, action, env)
        return ll, entropy



import os

import lightning as L
import torch

from omegaconf import DictConfig

# from rl4co.
from rl4co.utils.pylogger import get_pylogger

log = get_pylogger(__name__)


def get_lightning_device(lit_module: L.LightningModule) -> torch.device:
    """Get the device of the Lightning module before setup is called
    See device setting issue in setup https://github.com/Lightning-AI/lightning/issues/2638
    """
    try:
        if lit_module.trainer.strategy.root_device != lit_module.device:
            return lit_module.trainer.strategy.root_device
        return lit_module.device
    except Exception:
        return lit_module.device


def remove_key(config, key="wandb"):
    """Remove keys containing 'key`"""
    new_config = {}
    for k, v in config.items():
        if key in k:
            continue
        else:
            new_config[k] = v
    return new_config


def clean_hydra_config(
    config, keep_value_only=True, remove_keys="wandb", clean_cfg_path=True
):
    """Clean hydra config by nesting dictionary and cleaning values"""
    # Remove keys containing `remove_keys`
    if not isinstance(remove_keys, list):
        remove_keys = [remove_keys]
    for key in remove_keys:
        config = remove_key(config, key=key)

    new_config = {}
    # Iterate over config dictionary
    for key, value in config.items():
        # If key contains slash, split it and create nested dictionary recursively
        if "/" in key:
            keys = key.split("/")
            d = new_config
            for k in keys[:-1]:
                d = d.setdefault(k, {})
            d[keys[-1]] = value["value"] if keep_value_only else value
        else:
            new_config[key] = value["value"] if keep_value_only else value

    cfg = DictConfig(new_config)

    if clean_cfg_path:
        # Clean cfg_path recursively substituting root_dir with cwd
        root_dir = cfg.paths.root_dir

        def replace_dir_recursive(d, search, replace):
            for k, v in d.items():
                if isinstance(v, dict) or isinstance(v, DictConfig):
                    replace_dir_recursive(v, search, replace)
                elif isinstance(v, str):
                    if search in v:
                        d[k] = v.replace(search, replace)

        replace_dir_recursive(cfg, root_dir, os.getcwd())

    return cfg



import os

from pathlib import Path

from tqdm.auto import tqdm


def download_object_from_s3(
    session,
    *,
    bucket,
    key,
    version_id=None,
    filename,
    progress_bar=None,
    leave_pbar=False,
    pbar_position=1,
):
    """
    Download an object from S3 with a progress bar.


    Now actually download the object, with a progress bar to match.
    How this works:

    -   We take manual control of tqdm() using a ``with`` statement,
        see https://pypi.org/project/tqdm/#manual

    -   We set ``unit_scale=True`` so tqdm uses SI unit prefixes, and
       ``unit="B"`` means it adds a "B" as a suffix.  This means we get
        progress info like "14.5kB/s".

        (Note: the "B" is just a string; tqdm doesn't know these are
        bytes and doesn't care.)

    -   The Callback method on a boto3 S3 function is called
        periodically during the download with the number of bytes
        transferred.  We can use it to update the progress bar.

    From https://alexwlchan.net/2021/04/s3-progress-bars/
    """
    s3 = session.client("s3")

    # First get the size, so we know what tqdm is counting up to.
    # Theoretically the size could change between this HeadObject and starting
    # to download the file, but this would only affect the progress bar.
    kwargs = {"Bucket": bucket, "Key": key}

    if version_id is not None:
        kwargs["VersionId"] = version_id

    object_size = s3.head_object(**kwargs)["ContentLength"]

    if version_id is not None:
        ExtraArgs = {"VersionId": version_id}
    else:
        ExtraArgs = None
    filename = str(filename)

    if progress_bar is None:
        progress_bar = tqdm(
            total=object_size,
            unit="B",
            unit_scale=True,
            desc=filename,
            leave=leave_pbar,
            position=pbar_position,
        )

    # Make parent directories if not exists
    path = Path(filename)
    path.parent.absolute().mkdir(parents=True, exist_ok=True)

    with progress_bar as pbar:
        s3.download_file(
            Bucket=bucket,
            Key=key,
            ExtraArgs=ExtraArgs,
            Filename=filename,
            Callback=lambda bytes_transferred: pbar.update(bytes_transferred),
        )


def upload_file_to_s3(
    session,
    *,
    bucket,
    key,
    filename,
    progress_bar=None,
    leave_pbar=False,
    pbar_position=1,
):
    """
    Upload a file to S3 with a progress bar.

    From https://alexwlchan.net/2021/04/s3-progress-bars/
    """
    file_size = os.stat(filename).st_size
    filename = str(filename)

    s3 = session.client("s3")

    if progress_bar is None:
        progress_bar = tqdm(
            total=file_size,
            unit="B",
            unit_scale=True,
            desc=filename,
            leave=leave_pbar,
            position=pbar_position,
        )

    with progress_bar as pbar:
        s3.upload_file(
            Filename=filename,
            Bucket=bucket,
            Key=key,
            Callback=lambda bytes_transferred: pbar.update(bytes_transferred),
        )



import math

import torch
import torch.nn as nn
import torch.nn.functional as F

from rl4co.models.nn.utils import decode_probs


class SimpleAttention(nn.Module):
    """A generic attention module for a decoder in seq2seq"""

    def __init__(self, dim, use_tanh=False, C=10):
        super(SimpleAttention, self).__init__()
        self.use_tanh = use_tanh
        self.project_query = nn.Linear(dim, dim)
        self.project_ref = nn.Conv1d(dim, dim, 1, 1)
        self.C = C  # tanh exploration

        self.v = nn.Parameter(torch.FloatTensor(dim))
        self.v.data.uniform_(-(1.0 / math.sqrt(dim)), 1.0 / math.sqrt(dim))

    def forward(self, query, ref):
        """
        Args:
            query: is the hidden state of the decoder at the current
                time step. batch x dim
            ref: the set of hidden states from the encoder.
                sourceL x batch x hidden_dim
        """
        # ref is now [batch_size x hidden_dim x sourceL]
        ref = ref.permute(1, 2, 0)
        q = self.project_query(query).unsqueeze(2)  # batch x dim x 1
        e = self.project_ref(ref)  # batch_size x hidden_dim x sourceL
        # expand the query by sourceL
        # batch x dim x sourceL
        expanded_q = q.repeat(1, 1, e.size(2))
        # batch x 1 x hidden_dim
        v_view = self.v.unsqueeze(0).expand(expanded_q.size(0), len(self.v)).unsqueeze(1)
        # [batch_size x 1 x hidden_dim] * [batch_size x hidden_dim x sourceL]
        u = torch.bmm(v_view, F.tanh(expanded_q + e)).squeeze(1)
        if self.use_tanh:
            logits = self.C * F.tanh(u)
        else:
            logits = u
        return e, logits


class Decoder(nn.Module):
    def __init__(
        self,
        embedding_dim: int = 128,
        hidden_dim: int = 128,
        tanh_exploration: float = 10.0,
        use_tanh: bool = True,
        num_glimpses=1,
        mask_glimpses=True,
        mask_logits=True,
    ):
        super(Decoder, self).__init__()

        self.embedding_dim = embedding_dim
        self.hidden_dim = hidden_dim
        self.num_glimpses = num_glimpses
        self.mask_glimpses = mask_glimpses
        self.mask_logits = mask_logits
        self.use_tanh = use_tanh
        self.tanh_exploration = tanh_exploration

        self.lstm = nn.LSTMCell(embedding_dim, hidden_dim)
        self.pointer = SimpleAttention(hidden_dim, use_tanh=use_tanh, C=tanh_exploration)
        self.glimpse = SimpleAttention(hidden_dim, use_tanh=False)

    def update_mask(self, mask, selected):
        return mask.clone().scatter_(1, selected.unsqueeze(-1), True)

    def recurrence(self, x, h_in, prev_mask, prev_idxs, step, context):
        logit_mask = (
            self.update_mask(prev_mask, prev_idxs) if prev_idxs is not None else prev_mask
        )

        logits, h_out = self.calc_logits(
            x, h_in, logit_mask, context, self.mask_glimpses, self.mask_logits
        )

        # Calculate log_softmax for better numerical stability
        log_p = torch.log_softmax(logits, dim=1)
        probs = log_p.exp()

        if not self.mask_logits:
            probs[logit_mask] = 0.0

        return h_out, log_p, probs, logit_mask

    def calc_logits(
        self, x, h_in, logit_mask, context, mask_glimpses=None, mask_logits=None
    ):
        if mask_glimpses is None:
            mask_glimpses = self.mask_glimpses

        if mask_logits is None:
            mask_logits = self.mask_logits

        hy, cy = self.lstm(x, h_in)
        g_l, h_out = hy, (hy, cy)

        for i in range(self.num_glimpses):
            ref, logits = self.glimpse(g_l, context)
            # For the glimpses, only mask before softmax so we have always an L1 norm 1 readout vector
            if mask_glimpses:
                logits[logit_mask] = float("-inf")
            # [batch_size x h_dim x sourceL] * [batch_size x sourceL x 1] =
            # [batch_size x h_dim x 1]
            g_l = torch.bmm(ref, F.softmax(logits, dim=1).unsqueeze(2)).squeeze(2)
        _, logits = self.pointer(g_l, context)

        # Masking before softmax makes probs sum to one
        if mask_logits:
            logits[logit_mask] = float("-inf")

        return logits, h_out

    def forward(
        self,
        decoder_input,
        embedded_inputs,
        hidden,
        context,
        decode_type="sampling",
        eval_tours=None,
    ):
        """
        Args:
            decoder_input: The initial input to the decoder
                size is [batch_size x embedding_dim]. Trainable parameter.
            embedded_inputs: [sourceL x batch_size x embedding_dim]
            hidden: the prev hidden state, size is [batch_size x hidden_dim].
                Initially this is set to (enc_h[-1], enc_c[-1])
            context: encoder outputs, [sourceL x batch_size x hidden_dim]
        """

        batch_size = context.size(1)
        outputs = []
        selections = []
        steps = range(embedded_inputs.size(0))
        idxs = None
        mask = torch.zeros(
            embedded_inputs.size(1),
            embedded_inputs.size(0),
            dtype=torch.bool,
            device=embedded_inputs.device,
        )

        for i in steps:
            hidden, log_p, probs, mask = self.recurrence(
                decoder_input, hidden, mask, idxs, i, context
            )
            # select the next inputs for the decoder [batch_size x hidden_dim]
            idxs = (
                decode_probs(probs, mask, decode_type=decode_type)
                if eval_tours is None
                else eval_tours[:, i]
            )

            idxs = (
                idxs.detach()
            )  # Otherwise pytorch complains it want's a reward, todo implement this more properly?

            # Gather input embedding of selected
            decoder_input = torch.gather(
                embedded_inputs,
                0,
                idxs.contiguous()
                .view(1, batch_size, 1)
                .expand(1, batch_size, *embedded_inputs.size()[2:]),
            ).squeeze(0)

            # use outs to point to next object
            outputs.append(log_p)
            selections.append(idxs)
        return (torch.stack(outputs, 1), torch.stack(selections, 1)), hidden



import logging

from lightning.pytorch.utilities.rank_zero import rank_zero_only


def get_pylogger(name=__name__) -> logging.Logger:
    """Initializes multi-GPU-friendly python command line logger."""

    logger = logging.getLogger(name)

    # this ensures all logging levels get marked with the rank zero decorator
    # otherwise logs would get multiplied for each GPU process in multi-GPU setup
    logging_levels = (
        "debug",
        "info",
        "warning",
        "error",
        "exception",
        "fatal",
        "critical",
    )
    for level in logging_levels:
        setattr(logger, level, rank_zero_only(getattr(logger, level)))

    return logger



import math

import torch.nn as nn


class SkipConnection(nn.Module):
    def __init__(self, module):
        super(SkipConnection, self).__init__()
        self.module = module

    def forward(self, x):
        return x + self.module(x)


class Normalization(nn.Module):
    def __init__(self, embed_dim, normalization="batch"):
        super(Normalization, self).__init__()

        normalizer_class = {"batch": nn.BatchNorm1d, "instance": nn.InstanceNorm1d}.get(
            normalization, None
        )

        self.normalizer = normalizer_class(embed_dim, affine=True)

    def init_parameters(self):
        for name, param in self.named_parameters():
            stdv = 1.0 / math.sqrt(param.size(-1))
            param.data.uniform_(-stdv, stdv)

    def forward(self, x):
        if isinstance(self.normalizer, nn.BatchNorm1d):
            return self.normalizer(x.view(-1, x.size(-1))).view(*x.size())
        elif isinstance(self.normalizer, nn.InstanceNorm1d):
            return self.normalizer(x.permute(0, 2, 1)).permute(0, 2, 1)
        else:
            assert self.normalizer is None, "Unknown normalizer type"
            return x



from typing import Optional

import numpy as np
import torch

from tensordict.tensordict import TensorDict
from torchrl.data import (
    BoundedTensorSpec,
    CompositeSpec,
    UnboundedContinuousTensorSpec,
    UnboundedDiscreteTensorSpec,
)

from rl4co.envs.common.base import RL4COEnvBase
from rl4co.envs.common.utils import batch_to_scalar
from rl4co.utils.ops import gather_by_index, get_distance, get_tour_length


class MTSPEnv(RL4COEnvBase):
    """Multiple Traveling Salesman Problem environment
    At each step, an agent chooses to visit a city. A maximum of `num_agents` agents can be employed to visit the cities.
    The cost can be defined in two ways:
        - `minmax`: (default) the reward is the maximum of the path lengths of all the agents
        - `sum`: the cost is the sum of the path lengths of all the agents
    Reward is - cost, so the goal is to maximize the reward (minimize the cost).

    Args:
        num_loc: number of locations (cities) to visit
        min_loc: minimum value of the locations
        max_loc: maximum value of the locations
        min_num_agents: minimum number of agents
        max_num_agents: maximum number of agents
        cost_type: type of cost to use, either `minmax` or `sum`
        td_params: parameters for the TensorDict specs
    """

    name = "mtsp"

    def __init__(
        self,
        num_loc: int = 20,
        min_loc: float = 0,
        max_loc: float = 1,
        min_num_agents: int = 5,
        max_num_agents: int = 5,
        cost_type: str = "minmax",
        td_params: TensorDict = None,
        **kwargs,
    ):
        super().__init__(**kwargs)
        self.num_loc = num_loc
        self.min_loc = min_loc
        self.max_loc = max_loc
        self.min_num_agents = min_num_agents
        self.max_num_agents = max_num_agents
        self.cost_type = cost_type
        self._make_spec(td_params)

    @staticmethod
    def _step(td: TensorDict) -> TensorDict:
        # Initial variables
        is_first_action = batch_to_scalar(td["i"]) == 0
        current_node = td["action"]
        first_node = current_node if is_first_action else td["first_node"]

        # Get the locations of the current node and the previous node and the depot
        cur_loc = gather_by_index(td["locs"], current_node)
        prev_loc = gather_by_index(
            td["locs"], td["current_node"]
        )  # current_node is the previous node
        depot_loc = td["locs"][..., 0, :]

        # If current_node is the depot, then increment agent_idx
        cur_agent_idx = td["agent_idx"] + (current_node == 0).long()

        # Set not visited to 0 (i.e., we visited the node)
        available = td["action_mask"].scatter(
            -1, current_node[..., None].expand_as(td["action_mask"]), 0
        )
        # Available[..., 0] is the depot, which is always available unless:
        # - current_node is the depot
        # - agent_idx greater than num_agents -1
        available[..., 0] = torch.logical_and(
            current_node != 0, td["agent_idx"] < td["num_agents"] - 1
        )

        # We are done there are no unvisited locations except the depot
        done = torch.count_nonzero(available[..., 1:], dim=-1) == 0

        # If done is True, then we make the depot available again, so that it will be selected as the next node with prob 1
        available[..., 0] = torch.logical_or(done, available[..., 0])

        # If current agent is different from previous agent, then we have a new subtour and reset the length, otherwise we add the new distance
        current_length = torch.where(
            cur_agent_idx != td["agent_idx"],
            torch.zeros_like(td["current_length"]),
            td["current_length"] + get_distance(cur_loc, prev_loc),
        )

        # If done, we add the distance from the current_node to the depot as well
        current_length = torch.where(
            done, current_length + get_distance(cur_loc, depot_loc), current_length
        )

        # We update the max_subtour_length and reset the current_length
        max_subtour_length = torch.where(
            current_length > td["max_subtour_length"],
            current_length,
            td["max_subtour_length"],
        )

        # The reward is the negative of the max_subtour_length (minmax objective)
        reward = -max_subtour_length

        # The output must be written in a ``"next"`` entry
        return TensorDict(
            {
                "next": {
                    "locs": td["locs"],
                    "num_agents": td["num_agents"],
                    "max_subtour_length": max_subtour_length,
                    "current_length": current_length,
                    "agent_idx": cur_agent_idx,
                    "first_node": first_node,
                    "current_node": current_node,
                    "i": td["i"] + 1,
                    "action_mask": available,
                    "reward": reward,
                    "done": done,
                }
            },
            td.shape,
        )

    def _reset(self, td: Optional[TensorDict] = None, batch_size=None) -> TensorDict:
        # Initialize data
        if batch_size is None:
            batch_size = self.batch_size if td is None else td["locs"].shape[:-2]

        device = td.device if td is not None else self.device
        if td is None or td.is_empty():
            td = self.generate_data(batch_size=batch_size)

        # Keep track of the agent number to know when to stop
        agent_idx = torch.zeros((*batch_size,), dtype=torch.int64, device=device)

        # Make variable for max_subtour_length between subtours
        max_subtour_length = torch.zeros(batch_size, dtype=torch.float32, device=device)
        current_length = torch.zeros(batch_size, dtype=torch.float32, device=device)

        # Other variables
        current_node = torch.zeros((*batch_size,), dtype=torch.int64, device=device)
        available = torch.ones(
            (*batch_size, self.num_loc), dtype=torch.bool, device=device
        )  # 1 means not visited, i.e. action is allowed
        available[..., 0] = 0  # Depot is not available as first node
        i = torch.zeros((*batch_size,), dtype=torch.int64, device=device)

        return TensorDict(
            {
                "locs": td["locs"],  # depot is first node
                "num_agents": td["num_agents"],
                "max_subtour_length": max_subtour_length,
                "current_length": current_length,
                "agent_idx": agent_idx,
                "first_node": current_node,
                "current_node": current_node,
                "i": i,
                "action_mask": available,
            },
            batch_size=batch_size,
        )

    def _make_spec(self, td_params: TensorDict):
        """Make the observation and action specs from the parameters."""
        self.observation_spec = CompositeSpec(
            locs=BoundedTensorSpec(
                minimum=self.min_loc,
                maximum=self.max_loc,
                shape=(self.num_loc, 2),
                dtype=torch.float32,
            ),
            num_agents=UnboundedDiscreteTensorSpec(
                shape=(1),
                dtype=torch.int64,
            ),
            agent_idx=UnboundedDiscreteTensorSpec(
                shape=(1),
                dtype=torch.int64,
            ),
            current_length=UnboundedContinuousTensorSpec(
                shape=(1),
                dtype=torch.float32,
            ),
            max_subtour_length=UnboundedContinuousTensorSpec(
                shape=(1),
                dtype=torch.float32,
            ),
            first_node=UnboundedDiscreteTensorSpec(
                shape=(1),
                dtype=torch.int64,
            ),
            current_node=UnboundedDiscreteTensorSpec(
                shape=(1),
                dtype=torch.int64,
            ),
            i=UnboundedDiscreteTensorSpec(
                shape=(1),
                dtype=torch.int64,
            ),
            action_mask=UnboundedDiscreteTensorSpec(
                shape=(self.num_loc),
                dtype=torch.bool,
            ),
            shape=(),
        )
        self.input_spec = self.observation_spec.clone()
        self.action_spec = BoundedTensorSpec(
            shape=(1,),
            dtype=torch.int64,
            minimum=0,
            maximum=self.num_loc,
        )
        self.reward_spec = UnboundedContinuousTensorSpec()
        self.done_spec = UnboundedDiscreteTensorSpec(dtype=torch.bool)

    def get_reward(self, td, actions=None) -> TensorDict:
        # With minmax, get the maximum distance among subtours, calculated in the model
        if self.cost_type == "minmax":
            return td["reward"].squeeze(-1)

        # With distance, same as TSP
        elif self.cost_type == "distance":
            locs = td["locs"]
            locs_ordered = locs.gather(1, actions.unsqueeze(-1).expand_as(locs))
            return -get_tour_length(locs_ordered)

        else:
            raise ValueError(f"Cost type {self.cost_type} not supported")

    def generate_data(self, batch_size) -> TensorDict:
        batch_size = [batch_size] if isinstance(batch_size, int) else batch_size

        # Initialize the locations (including the depot which is always the first node)
        locs = (
            torch.FloatTensor(*batch_size, self.num_loc, 2)
            .uniform_(self.min_loc, self.max_loc)
            .to(self.device)
        )

        # Initialize the num_agents: either fixed or random integer between min and max
        if self.min_num_agents == self.max_num_agents:
            num_agents = (
                torch.ones(batch_size, dtype=torch.int64, device=self.device)
                * self.min_num_agents
            )
        else:
            num_agents = torch.randint(
                self.min_num_agents,
                self.max_num_agents,
                size=batch_size,
                device=self.device,
            )

        return TensorDict(
            {
                "locs": locs,
                "num_agents": num_agents,
            },
            batch_size=batch_size,
        )

    @staticmethod
    def render(td, actions=None, ax=None):
        import matplotlib.pyplot as plt

        from matplotlib import colormaps

        def discrete_cmap(num, base_cmap="nipy_spectral"):
            """Create an N-bin discrete colormap from the specified input map"""
            base = colormaps[base_cmap]
            color_list = base(np.linspace(0, 1, num))
            cmap_name = base.name + str(num)
            return base.from_list(cmap_name, color_list, num)

        if actions is None:
            actions = td.get("action", None)
        # if batch_size greater than 0 , we need to select the first batch element
        if td.batch_size != torch.Size([]):
            td = td[0]
            actions = actions[0]

        num_agents = td["num_agents"]
        locs = td["locs"]
        cmap = discrete_cmap(num_agents, "rainbow")

        fig, ax = plt.subplots()

        # Add depot action = 0 to before first action and after last action
        actions = torch.cat(
            [
                torch.zeros(1, dtype=torch.int64),
                actions,
                torch.zeros(1, dtype=torch.int64),
            ]
        )

        # Make list of colors from matplotlib
        for i, loc in enumerate(locs):
            if i == 0:
                # depot
                marker = "s"
                color = "g"
                label = "Depot"
                markersize = 10
            else:
                # normal location
                marker = "o"
                color = "tab:blue"
                label = "Cities"
                markersize = 8
            if i > 1:
                label = ""

            ax.plot(
                loc[0],
                loc[1],
                color=color,
                marker=marker,
                markersize=markersize,
                label=label,
            )

        # Plot the actions in order
        agent_idx = 0
        for i in range(len(actions)):
            if actions[i] == 0:
                agent_idx += 1
            color = cmap(num_agents - agent_idx)

            from_node = actions[i]
            to_node = (
                actions[i + 1] if i < len(actions) - 1 else actions[0]
            )  # last goes back to depot
            from_loc = td["locs"][from_node]
            to_loc = td["locs"][to_node]
            ax.plot([from_loc[0], to_loc[0]], [from_loc[1], to_loc[1]], color=color)
            ax.annotate(
                "",
                xy=(to_loc[0], to_loc[1]),
                xytext=(from_loc[0], from_loc[1]),
                arrowprops=dict(arrowstyle="->", color=color),
                annotation_clip=False,
            )

        # Legend
        handles, labels = ax.get_legend_handles_labels()
        ax.legend(handles, labels)
        ax.set_title("mTSP")
        ax.set_xlabel("x-coordinate")
        ax.set_ylabel("y-coordinate")
        plt.show()




from typing import Union

from rl4co.envs.common.base import RL4COEnvBase
from rl4co.models.rl import REINFORCE
from rl4co.models.rl.reinforce.baselines import REINFORCEBaseline
from rl4co.models.zoo.mdam.policy import MDAMPolicy


class MDAM(REINFORCE):
    """ Multi-Decoder Attention Model (MDAM) is a model
    to train multiple diverse policies, which effectively increases the chance of finding 
    good solutions compared with existing methods that train only one policy.
    Reference link: https://arxiv.org/abs/2012.10638; 
    Implementation reference: https://github.com/liangxinedu/MDAM.

    Args:
        env: Environment to use for the algorithm
        policy: Policy to use for the algorithm
        baseline: REINFORCE baseline. Defaults to rollout (1 epoch of exponential, then greedy rollout baseline)
        policy_kwargs: Keyword arguments for policy
        baseline_kwargs: Keyword arguments for baseline
        **kwargs: Keyword arguments passed to the superclass
    """

    def __init__(
            self, 
            env: RL4COEnvBase, 
            policy: MDAMPolicy = None, 
            baseline: Union[REINFORCEBaseline, str] = "rollout", 
            policy_kwargs={},
            baseline_kwargs={},
            **kwargs
        ):
        if policy is None:
            policy = MDAMPolicy(env.name, **policy_kwargs) 

        super().__init__(env, policy, baseline, baseline_kwargs, **kwargs)


from typing import Optional

import numpy as np
import torch

from tensordict.tensordict import TensorDict
from torchrl.data import (
    BoundedTensorSpec,
    CompositeSpec,
    UnboundedContinuousTensorSpec,
    UnboundedDiscreteTensorSpec,
)

from rl4co.envs.dpp import DPPEnv
from rl4co.utils.pylogger import get_pylogger

log = get_pylogger(__name__)


class MDPPEnv(DPPEnv):
    """Multiple decap placement problem (mDPP) environment
    This is a modified version of the DPP environment where we allow multiple probing ports
    The reward can be calculated as:
        - minmax: min of the max of the decap scores
        - meansum: mean of the sum of the decap scores
    The minmax is more challenging as it requires to find the best decap location for the worst case

    Args:
        num_probes_min: minimum number of probes
        num_probes_max: maximum number of probes
        reward_type: reward type, either minmax or meansum
        td_params: TensorDict parameters
    """

    name = "mdpp"

    def __init__(
        self,
        *,
        num_probes_min: int = 2,
        num_probes_max: int = 5,
        reward_type: str = "minmax",
        td_params: TensorDict = None,
        **kwargs,
    ):
        super().__init__(**kwargs)
        self.num_probes_min = num_probes_min
        self.num_probes_max = num_probes_max
        assert reward_type in [
            "minmax",
            "meansum",
        ], "reward_type must be minmax or meansum"
        self.reward_type = reward_type
        self._make_spec(td_params)

    def _step(self, td: TensorDict) -> TensorDict:
        # Step function is the same as DPPEnv, only masking changes
        return super()._step(td)

    def _reset(self, td: Optional[TensorDict] = None, batch_size=None) -> TensorDict:
        # Reset function is the same as DPPEnv, only masking changes due to probes
        td_reset = super()._reset(td, batch_size=batch_size)

        # Action mask is 0 if both action_mask (e.g. keepout) and probe are 0
        action_mask = torch.logical_and(td_reset["action_mask"], ~td_reset["probe"])
        # Keepout regions are the inverse of action_mask
        td_reset.set_("keepout", ~td_reset["action_mask"])
        td_reset.set_("action_mask", action_mask)
        return td_reset

    def _make_spec(self, td_params):
        """Make the observation and action specs from the parameters"""
        self.observation_spec = CompositeSpec(
            locs=BoundedTensorSpec(
                minimum=self.min_loc,
                maximum=self.max_loc,
                shape=(self.size**2, 2),
                dtype=torch.float32,
            ),
            probe=UnboundedDiscreteTensorSpec(
                shape=(self.size**2),
                dtype=torch.bool,
            ),  # probe is a boolean of multiple locations (1=probe, 0=not probe)
            keepout=UnboundedDiscreteTensorSpec(
                shape=(self.size**2),
                dtype=torch.bool,
            ),
            i=UnboundedDiscreteTensorSpec(
                shape=(1),
                dtype=torch.int64,
            ),
            action_mask=UnboundedDiscreteTensorSpec(
                shape=(self.size**2),
                dtype=torch.bool,
            ),
            shape=(),
        )
        self.input_spec = self.observation_spec.clone()
        self.action_spec = BoundedTensorSpec(
            shape=(1,),
            dtype=torch.int64,
            minimum=0,
            maximum=self.size**2,
        )
        self.reward_spec = UnboundedContinuousTensorSpec(shape=(1,))
        self.done_spec = UnboundedDiscreteTensorSpec(shape=(1,), dtype=torch.bool)

    def get_reward(self, td, actions):
        """We call the reward function with the final sequence of actions to get the reward
        Calling per-step would be very time consuming due to decap simulation
        """
        # We do the operation in a batch
        if len(td.batch_size) == 0:
            td = td.unsqueeze(0)
            actions = actions.unsqueeze(0)

        # Reward calculation is expensive since we need to run decap simulation (not vectorizable)
        reward = torch.stack(
            [
                self._single_env_reward(td_single, action)
                for td_single, action in zip(td, actions)
            ]
        )
        return reward

    def _single_env_reward(self, td, actions):
        """Get reward for single environment. We"""

        list_probe = torch.nonzero(td["probe"]).squeeze()
        scores = torch.zeros_like(list_probe, dtype=torch.float32)
        for i, probe in enumerate(list_probe):
            # Get the decap scores for the probe location
            scores[i] = self._decap_simulator(probe, actions)
        # If minmax, return min of max decap scores else mean
        return scores.min() if self.reward_type == "minmax" else scores.mean()

    def generate_data(self, batch_size):
        """
        Generate initial observations for the environment with locations, probe, and action mask
        Action_mask eliminates the keepout regions and the probe location, and is updated to eliminate placed decaps
        """

        m = n = self.size
        # if int, convert to list and make it a batch for easier generation
        batch_size = [batch_size] if isinstance(batch_size, int) else batch_size
        batched = len(batch_size) > 0
        bs = [1] if not batched else batch_size

        # Create a list of locs on a grid
        locs = torch.meshgrid(torch.arange(m), torch.arange(n))
        locs = torch.stack(locs, dim=-1).reshape(-1, 2)
        # normalize the locations by the number of rows and columns
        locs = locs / torch.tensor([m, n], dtype=torch.float)
        locs = locs[None].expand(*bs, -1, -1)

        # Create available mask
        available = torch.ones((*bs, m * n), dtype=torch.bool)

        # Sample probe location from m*n
        probe = torch.randint(m * n, size=(*bs, 1))
        available.scatter_(1, probe, False)

        # Sample probe locatins
        num_probe = torch.randint(
            self.num_probes_min,
            self.num_probes_max,
            size=(*bs, 1),
        )
        probe = [torch.randperm(m * n)[:p] for p in num_probe]
        probes = torch.zeros((*bs, m * n), dtype=torch.bool)
        for i, (a, p) in enumerate(zip(available, probe)):
            available[i] = a.scatter(0, p, False)
            probes[i] = probes[i].scatter(0, p, True)

        # Sample keepout locations from m*n except probe
        num_keepout = torch.randint(
            self.num_keepout_min,
            self.num_keepout_max,
            size=(*bs, 1),
        )
        keepouts = [torch.randperm(m * n)[:k] for k in num_keepout]
        for i, (a, k) in enumerate(zip(available, keepouts)):
            available[i] = a.scatter(0, k, False)

        return TensorDict(
            {
                "locs": locs if batched else locs.squeeze(0),
                "probe": probes if batched else probes.squeeze(0),
                "action_mask": available if batched else available.squeeze(0),
            },
            batch_size=batch_size,
        )

    def render(self, td, actions=None, ax=None, legend=True, settings=None):
        """Plot a grid of squares representing the environment.
        The keepout regions are the action_mask - decaps - probe
        """

        import matplotlib.pyplot as plt

        from matplotlib.lines import Line2D
        from matplotlib.patches import Annulus, Rectangle, RegularPolygon

        if settings is None:
            settings = {
                "available": {"color": "white", "label": "available"},
                "keepout": {"color": "grey", "label": "keepout"},
                "probe": {"color": "tab:red", "label": "probe"},
                "decap": {"color": "tab:blue", "label": "decap"},
            }

        def draw_capacitor(ax, x, y, color="black"):
            # Backgrund rectangle: same as color but with alpha=0.5
            ax.add_patch(Rectangle((x, y), 1, 1, color=color, alpha=0.5))

            # Create the plates of the capacitor
            plate_width, plate_height = (
                0.3,
                0.1,
            )  # Width and height switched to make vertical
            plate_gap = 0.2
            plate1 = Rectangle(
                (x + 0.5 - plate_width / 2, y + 0.5 - plate_height - plate_gap / 2),
                plate_width,
                plate_height,
                color=color,
            )
            plate2 = Rectangle(
                (x + 0.5 - plate_width / 2, y + 0.5 + plate_gap / 2),
                plate_width,
                plate_height,
                color=color,
            )

            # Add the plates to the axes
            ax.add_patch(plate1)
            ax.add_patch(plate2)

            # Add connection lines (wires)
            line_length = 0.2
            line1 = Line2D(
                [x + 0.5, x + 0.5],
                [
                    y + 0.5 - plate_height - plate_gap / 2 - line_length,
                    y + 0.5 - plate_height - plate_gap / 2,
                ],
                color=color,
            )
            line2 = Line2D(
                [x + 0.5, x + 0.5],
                [
                    y + 0.5 + plate_height + plate_gap / 2,
                    y + 0.5 + plate_height + plate_gap / 2 + line_length,
                ],
                color=color,
            )

            # Add the lines to the axes
            ax.add_line(line1)
            ax.add_line(line2)

        def draw_probe(ax, x, y, color="black"):
            # Backgrund rectangle: same as color but with alpha=0.5
            ax.add_patch(Rectangle((x, y), 1, 1, color=color, alpha=0.5))
            ax.add_patch(Annulus((x + 0.5, y + 0.5), (0.2, 0.2), 0.1, color=color))

        def draw_keepout(ax, x, y, color="black"):
            # Backgrund rectangle: same as color but with alpha=0.5
            ax.add_patch(Rectangle((x, y), 1, 1, color=color, alpha=0.5))
            ax.add_patch(
                RegularPolygon(
                    (x + 0.5, y + 0.5), numVertices=6, radius=0.45, color=color
                )
            )

        size = self.size
        td = td.detach().cpu()
        # if batch_size greater than 0 , we need to select the first batch element
        if td.batch_size != torch.Size([]):
            td = td[0]

        if actions is None:
            actions = td.get("action", None)

        # Transform actions from idx to one-hot
        decaps = torch.zeros(size**2)
        decaps.scatter_(0, actions, 1)
        decaps = decaps.reshape(size, size)

        keepout = ~td["action_mask"].reshape(size, size)
        probes = td["probe"].reshape(size, size)

        if ax is None:
            _, ax = plt.subplots(1, 1, figsize=(6, 6))

        grid = np.meshgrid(np.arange(0, size), np.arange(0, size))
        grid = np.stack(grid, axis=-1)

        xdim, ydim = grid.shape[0], grid.shape[1]
        # ax.imshow(np.zeros((xdim, ydim)), cmap="gray")

        ax.set_xlim(0, xdim)
        ax.set_ylim(0, ydim)

        for i in range(xdim):
            for j in range(ydim):
                x, y = grid[i, j, 0], grid[i, j, 1]

                if decaps[i, j] == 1:
                    draw_capacitor(ax, x, y, color=settings["decap"]["color"])
                elif probes[i, j] == 1:
                    draw_probe(ax, x, y, color=settings["probe"]["color"])
                elif keepout[i, j] == 1:
                    draw_keepout(ax, x, y, color=settings["keepout"]["color"])

        ax.grid(
            which="major", axis="both", linestyle="-", color="k", linewidth=1, alpha=0.5
        )
        # set 10 ticks
        ax.set_xticks(np.arange(0, xdim, 1))
        ax.set_yticks(np.arange(0, ydim, 1))

        # Invert y axis
        ax.invert_yaxis()

        # # Add legend
        if legend:
            colors = [settings[k]["color"] for k in settings.keys()]
            labels = [settings[k]["label"] for k in settings.keys()]
            handles = [
                plt.Rectangle(
                    (0, 0), 1, 1, color=c, edgecolor="k", linestyle="-", linewidth=1
                )
                for c in colors
            ]
            ax.legend(
                handles,
                [label for label in labels],
                ncol=len(colors),
                loc="upper center",
                bbox_to_anchor=(0.5, 1.1),
            )



import torch
import torch.nn as nn

from .decoder import SimpleAttention
from .encoder import Encoder


class CriticNetworkLSTM(nn.Module):
    """Useful as a baseline in REINFORCE updates"""

    def __init__(
        self,
        embedding_dim,
        hidden_dim,
        n_process_block_iters,
        tanh_exploration,
        use_tanh,
    ):
        super(CriticNetworkLSTM, self).__init__()

        self.hidden_dim = hidden_dim
        self.n_process_block_iters = n_process_block_iters

        self.encoder = Encoder(embedding_dim, hidden_dim)

        self.process_block = SimpleAttention(
            hidden_dim, use_tanh=use_tanh, C=tanh_exploration
        )
        self.sm = nn.Softmax(dim=1)
        self.decoder = nn.Sequential(
            nn.Linear(hidden_dim, hidden_dim), nn.ReLU(), nn.Linear(hidden_dim, 1)
        )

    def forward(self, inputs):
        """
        Args:
            inputs: [embedding_dim x batch_size x sourceL] of embedded inputs
        """
        inputs = inputs.transpose(0, 1).contiguous()

        encoder_hx = (
            self.encoder.init_hx.unsqueeze(0).repeat(inputs.size(1), 1).unsqueeze(0)
        )
        encoder_cx = (
            self.encoder.init_cx.unsqueeze(0).repeat(inputs.size(1), 1).unsqueeze(0)
        )

        # encoder forward pass
        enc_outputs, (enc_h_t, enc_c_t) = self.encoder(inputs, (encoder_hx, encoder_cx))

        # grab the hidden state and process it via the process block
        process_block_state = enc_h_t[-1]
        for i in range(self.n_process_block_iters):
            ref, logits = self.process_block(process_block_state, enc_outputs)
            process_block_state = torch.bmm(ref, self.sm(logits).unsqueeze(2)).squeeze(2)
        # produce the final scalar output
        out = self.decoder(process_block_state)
        return out
